/*
  Name: aC_Joystick.HPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - C (Control) Subsystem - Joystick abstraction
*/

#include "aC_Joystick.hpp"
#include <math.h>

__fastcall Aegis::C::Joystick::Joystick(unsigned int joy) {
    int num = joyGetNumDevs();
    if (joy < num) {
        this->joystick = joy;
        this->quit = false;
        
        joyGetDevCaps( this->joystick, &this->joycaps, sizeof(JOYCAPS) );

        DWORD dw = 0;
        this->hThread = CreateThread( NULL, 65536, this->PollingThread,
                                      this, 0, &dw);
    } else {
        // nothing appliable
        this->joystick = 0;
        this->hThread = NULL;
        ZeroMemory( &this->joyinfo, sizeof(JOYINFOEX) );
    };
    
    InitializeCriticalSection(&this->cs);

    ZeroMemory( &this->imm_on_table, sizeof(unsigned int) * 32 );
    ZeroMemory( &this->inc_table, sizeof(unsigned int) * 32 );
    ZeroMemory( &this->time_table, sizeof(unsigned int) * 32 );
    ZeroMemory( &this->uptime_table, sizeof(unsigned int) * 32 );
    
    this->DeadZone( 0.1f ); // hide shaking
    this->FullZone( 0.9f ); // force full range
    
    return;
};

__fastcall Aegis::C::Joystick::~Joystick() {
    this->quit = true;

    if (this->hThread)
        WaitForSingleObject( this->hThread, INFINITE );
    
    DeleteCriticalSection(&this->cs);
    return;
};

DWORD WINAPI Aegis::C::Joystick::PollingThread(LPVOID lpvoid) {
    Aegis::C::Joystick *This = (Aegis::C::Joystick *)lpvoid;

    unsigned int temp = 0;
    JOYINFOEX nji = { sizeof(JOYINFOEX), 
                      JOY_RETURNX | JOY_RETURNY | JOY_RETURNBUTTONS };
                      
    if (This->joycaps.wCaps & JOYCAPS_HASZ) nji.dwFlags |= JOY_RETURNZ;
    if (This->joycaps.wCaps & JOYCAPS_HASR) nji.dwFlags |= JOY_RETURNR;
    if (This->joycaps.wCaps & JOYCAPS_HASU) nji.dwFlags |= JOY_RETURNU;
    if (This->joycaps.wCaps & JOYCAPS_HASV) nji.dwFlags |= JOY_RETURNV;
    if (This->joycaps.wCaps & JOYCAPS_HASPOV) {
        if (This->joycaps.wCaps & JOYCAPS_POV4DIR) nji.dwFlags |= JOY_RETURNPOV;
        if (This->joycaps.wCaps & JOYCAPS_POVCTS)  nji.dwFlags |= JOY_RETURNPOVCTS;
    };

    while(!This->quit) {
        temp = GetTickCount();

        Sleep(50); //Sleep(100); // joystick polling is expensive
        joyGetPosEx( This->joystick, &nji );

        temp = GetTickCount() - temp;

        EnterCriticalSection( &This->cs );
        
        DWORD prevBtn = This->joyinfo.dwButtons;
        This->joyinfo = nji;
        
        for (int i = 0; i < 32; ++i) {
                if (This->IsDown(i)) {
                    This->time_table[i] += temp;
                    This->uptime_table[i] = 0; // auto reset - only uptime
                    
                    if (!(prevBtn & (1 << i))) {
                        ++This->inc_table[i];
                        This->imm_on_table[i] = !This->imm_on_table[i];
                    };
                } else {
                    This->uptime_table[i] += temp;
                };    
            };//*/
        
        LeaveCriticalSection( &This->cs );
    };
    
    return 0;
};

__fastcall bool Aegis::C::Joystick::DeadZone(float dz) {
    if (dz < 0.0f || dz > 0.4f) return false;
    
    this->deadzone = dz;    
    return true;
};
__fastcall bool Aegis::C::Joystick::FullZone(float fz) {
    if (fz < 0.5f || fz > 1.0f) return false;

    this->fullzone = fz;    
    return true;
};
__fastcall float Aegis::C::Joystick::Adjust(float axis) {
    // adjust deadzone
    if (axis > this->deadzone) axis -= this->deadzone;
    else
    if (axis < -this->deadzone) axis += this->deadzone;
    else
        axis = 0.0f;

    // adjust fullzone - change slope
    axis *= 1.0f / (this->fullzone - this->deadzone);

    // need flat fullzone
    if (axis < -1.0f) axis = -1.0f;
    if (axis >  1.0f) axis =  1.0f;
    
    return axis;
};

__fastcall float Aegis::C::Joystick::GetX() {
    unsigned int mid = (this->joycaps.wXmax + this->joycaps.wXmin) >> 1;
    unsigned int len = (this->joycaps.wXmax - this->joycaps.wXmin) >> 1;
    int pos = this->joyinfo.dwXpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetY() {
    unsigned int mid = (this->joycaps.wYmax + this->joycaps.wYmin) >> 1;
    unsigned int len = (this->joycaps.wYmax - this->joycaps.wYmin) >> 1;
    int pos = this->joyinfo.dwYpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetZ() {
    unsigned int mid = (this->joycaps.wZmax + this->joycaps.wZmin) >> 1;
    unsigned int len = (this->joycaps.wZmax - this->joycaps.wZmin) >> 1;
    int pos = this->joyinfo.dwZpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetR() {
    unsigned int mid = (this->joycaps.wRmax + this->joycaps.wRmin) >> 1;
    unsigned int len = (this->joycaps.wRmax - this->joycaps.wRmin) >> 1;
    int pos = this->joyinfo.dwRpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetU() {
    unsigned int mid = (this->joycaps.wUmax + this->joycaps.wUmin) >> 1;
    unsigned int len = (this->joycaps.wUmax - this->joycaps.wUmin) >> 1;
    int pos = this->joyinfo.dwUpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetV() {
    unsigned int mid = (this->joycaps.wVmax + this->joycaps.wVmin) >> 1;
    unsigned int len = (this->joycaps.wVmax - this->joycaps.wVmin) >> 1;
    int pos = this->joyinfo.dwVpos - mid;

    return this->Adjust((float)pos / (float)len);
};

__fastcall float Aegis::C::Joystick::GetPOV() {
    if (this->joyinfo.dwPOV == JOY_POVCENTERED)
        // Use rather < 0.0f or IsPOVCentered function as this may change
        return -1.0f;
    else
        // This is safe for both CTS and discrete POVs.
        return (float)this->joyinfo.dwPOV / 100.0f;
};

// this handlers are set up by application
__fastcall int Aegis::C::Joystick::MovementHandler(unsigned int axis){ return 0; };
__fastcall int Aegis::C::Joystick::POVHandler() { return 0; };
__fastcall int Aegis::C::Joystick::UpDownHandler( unsigned int button,
                                                  bool down ) { return 0; };
__fastcall int Aegis::C::Joystick::OnOffHandler( unsigned int button,
                                                 bool down ) { return 0; };

// counter functions        
__fastcall unsigned int Aegis::C::Joystick::GetHits( unsigned int button,
                                                     unsigned int max ) {
    EnterCriticalSection( &this->cs );

    int n;
    if (!max) {
        n = this->inc_table[ button ];
    } else {    
        n = ( this->inc_table[ button ] <? max );
        this->inc_table[ button ] -= n;
    };
    
    LeaveCriticalSection( &this->cs );
    return n;
};

__fastcall unsigned int Aegis::C::Joystick::GetTime( unsigned int button,
                                                     unsigned int max ) {
    EnterCriticalSection( &this->cs );
    
    int n;
    if (!max) {
        n = this->time_table[ button ];
    } else {    
        n = ( this->time_table[ button ] <? max );
        this->time_table[ button ] -= n;
    };
    
    LeaveCriticalSection( &this->cs );
    return n;
};

__fastcall unsigned int Aegis::C::Joystick::GetUpTime( unsigned int button,
                                                       unsigned int max ) {
    EnterCriticalSection( &this->cs );
    
    int n;
    if (!max) {
        n = this->uptime_table[ button ];
    } else {    
        n = ( this->uptime_table[ button ] <? max );
        this->uptime_table[ button ] -= n;
    };
    
    LeaveCriticalSection( &this->cs );
    return n;
};

