/*
  Name: aF_Bitmap.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - F (File Formats) - Bitmap (aBitmap) base class  
*/

#include "aF_Bitmap.hpp"
#include <stdio.h>
#include <zlib.h>
#include <math.h>
#include <new>

static const char *BitmapMagic = "aBitmap";

__fastcall Aegis::F::Bitmap::Bitmap() {
    ZeroMemory( this, sizeof(Aegis::F::Bitmap) );
    strcpy( this->magic, BitmapMagic );
    
    return;
};    

__fastcall Aegis::F::Bitmap::Bitmap(long x,long y) {
    ZeroMemory( this, sizeof(Aegis::F::Bitmap) );
    strcpy( this->magic, BitmapMagic );

    this->sizex = x;
    this->sizey = y;
    this->rgba = new(std::nothrow) Aegis::M2::RGBA[ this->sizex * this->sizey ];
    
    return;
};

__fastcall Aegis::F::Bitmap::Bitmap( const Aegis::F::Bitmap& bm ) {
    CopyMemory( this, &bm, sizeof(Aegis::F::Bitmap) );

    int s = this->sizex * this->sizey;
    this->rgba = new(std::nothrow) Aegis::M2::RGBA[s];

    if (this->rgba) CopyMemory( this->rgba, bm.rgba, s * sizeof(Aegis::M2::RGBA) );
    return;
};

__fastcall Aegis::F::Bitmap::~Bitmap() {
    if (this->rgba) delete[] this->rgba;
    
    return;
};

__fastcall int Aegis::F::Bitmap::Load( char *name ) {
    HANDLE hFile = CreateFile( name, GENERIC_READ, FILE_SHARE_READ, NULL,
                               OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    if (hFile == INVALID_HANDLE_VALUE) return GetLastError();
    
    DWORD readBytes = 0;
    unsigned int header_size = sizeof(Aegis::F::Bitmap)
                             - sizeof(Aegis::M2::RGBA *);

    Aegis::F::Bitmap dumpBitmap;
    if (!ReadFile( hFile, &dumpBitmap, header_size, &readBytes, NULL )) {
        // problem with reading
        CloseHandle( hFile );
        return GetLastError();
    };
    
    // check file structure
    if (strcmpi( dumpBitmap.magic, BitmapMagic )) return -1;
    
    if (readBytes < header_size) return -2;

    // decompress
    int cmperr = 0;
    Aegis::M2::RGBA *buffer = new(std::nothrow)
        Aegis::M2::RGBA[ dumpBitmap.sizex * dumpBitmap.sizey ];
         
    if (!buffer) return -3; 

    SetFilePointer( hFile, header_size, NULL, FILE_BEGIN );

    switch( dumpBitmap.compression ) {
        case 0: { // no compression
            int readSize = dumpBitmap.sizex * dumpBitmap.sizey
                         * sizeof(Aegis::M2::RGBA);
            if (!ReadFile( hFile, buffer, readSize, &readBytes, NULL ))
                cmperr = -5;
        } break;

        case 1: { // zlib direct
            int readSize = GetFileSize( hFile, NULL ) - header_size;
            char *temp = new(std::nothrow) char[ readSize ];
            if (!temp) return -3; 
            
            if (!ReadFile( hFile, temp, readSize, &readBytes, NULL ))
                cmperr = -5;
            
            uLongf destLen = dumpBitmap.sizex * dumpBitmap.sizey
                           * sizeof(Aegis::M2::RGBA);
            int ret = uncompress( (Bytef*) buffer, &destLen,
                                  (const Bytef*) temp, readSize );
            
            if (ret == Z_MEM_ERROR) return -3;
            if (ret == Z_BUF_ERROR) return -6;
            if (ret == Z_DATA_ERROR) return -7;
            if (ret != Z_OK) return -8;
                
            delete temp;
        } break;
        
        // TODO: Add more compressions
     
        
        default:
            cmperr = -9; // unknown compression
    };    
    
    if (cmperr) {
        delete[] buffer;
        CloseHandle( hFile );
        return cmperr;
    };

    // finalize
    if (this->rgba) delete[] this->rgba;
    this->rgba = buffer;
    
    this->sizex = dumpBitmap.sizex;
    this->sizey = dumpBitmap.sizey;
    this->compression = dumpBitmap.compression;
    this->border = dumpBitmap.border;
    
    this->p3 = dumpBitmap.p3;
    this->p4 = dumpBitmap.p4;
    
    CloseHandle( hFile );
    return 0;
};
    
__fastcall int Aegis::F::Bitmap::Save( char *name, int cmp ) {
    HANDLE hFile = CreateFile( name, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if (hFile == INVALID_HANDLE_VALUE) return GetLastError();

    this->compression = cmp;

    DWORD writtenBytes = 0;
    unsigned int header_size = sizeof(Aegis::F::Bitmap)
                             - sizeof(Aegis::M2::RGBA *);

    if (!WriteFile( hFile, this, header_size, &writtenBytes, NULL )) {
        // problem with writting
        CloseHandle( hFile );
        return GetLastError();
    };
    
    void *bufferToWrite;
    unsigned int writeBytes = 0;
    switch( this->compression ) {
        case 0: { // no compression
            bufferToWrite = this->rgba;
            writeBytes = sizeof(Aegis::M2::RGBA)
                       * this->sizex * this->sizey;
        } break;    
        case 1: { // zlib direct
            writeBytes = compressBound(sizeof(Aegis::M2::RGBA)
                                       * this->sizex * this->sizey);
            bufferToWrite = GlobalAlloc( GPTR, writeBytes );
            if (!bufferToWrite) {
                CloseHandle( hFile );
                return -3;
            };
            
            int ret = compress2( (Bytef*)bufferToWrite, (uLongf*)&writeBytes,
                                 (Bytef*)this->rgba, sizeof(Aegis::M2::RGBA)
                                 * this->sizex * this->sizey, 9);

            if (ret != Z_OK) CloseHandle( hFile );
            if (ret == Z_MEM_ERROR) return -3;
            if (ret == Z_BUF_ERROR) return -6;
            if (ret != Z_OK) return -8;
        } break;
/*        case 2: { // zlib repacked
            int repackedBytes = sizeof(Aegis::M2::RGBA)
                              * this->sizex * this->sizey;
            unsigned char *repackedBuffer = (unsigned char*)
                                            GlobalAlloc( GPTR, repackedBytes );
            if (!repackedBuffer) {
                CloseHandle( hFile );
                return -3;
            };
                              
            writeBytes = compressBound(repackedBytes);
            bufferToWrite = GlobalAlloc( GPTR, writeBytes );

            if (!bufferToWrite) {
                GlobalFree( repackedBuffer );
                CloseHandle( hFile );
                return -3;
            };
            
            // repack
            int block = this->sizex * this->sizey;
            for (int y = 0; y < this->sizey; ++y)
                for (int x = 0; x < this->sizex; ++x) {
                    int off = y*this->sizex + x;
                    repackedBuffer[ 0*block + off ] = this->rgba[off].r;
                    repackedBuffer[ 1*block + off ] = this->rgba[off].g;
                    repackedBuffer[ 2*block + off ] = this->rgba[off].b;
                    repackedBuffer[ 3*block + off ] = this->rgba[off].a;
                };    
            
            // compress
            int ret = compress2( (Bytef*)bufferToWrite, (uLongf*)&writeBytes,
                                 (Bytef*)repackedBuffer, repackedBytes, 9);

            GlobalFree( repackedBuffer );
            
            if (ret != Z_OK) CloseHandle( hFile );
            if (ret == Z_MEM_ERROR) return -3;
            if (ret == Z_BUF_ERROR) return -6;
            if (ret != Z_OK) return -8;
        } break;*/
        default:
            CloseHandle( hFile );
            return -1;
    };
    
    if (!WriteFile( hFile, bufferToWrite, writeBytes, &writtenBytes, NULL )) {
        // problem with writting
        if (this->compression) GlobalFree(bufferToWrite);
        CloseHandle( hFile );
        
        return GetLastError();
    };
    
    if (this->compression) GlobalFree(bufferToWrite);

    CloseHandle( hFile );
    return 0;
};

__fastcall int Aegis::F::Bitmap::LoadHBITMAP( HDC hDC, HBITMAP hBitmap ) {
    BITMAP bm;
    GetObject( hBitmap, sizeof(BITMAP), &bm );

    if (bm.bmWidth != this->sizex || bm.bmHeight != this->sizey) {
        if (this->rgba) delete[] this->rgba;
        this->sizex = bm.bmWidth;
        this->sizey = bm.bmHeight;
        rgba = new Aegis::M2::RGBA[ this->sizex * this->sizey ];
    }

    BITMAPINFOHEADER bh = {
        sizeof(BITMAPINFOHEADER),
        bm.bmWidth,
        bm.bmHeight,
        1,
        32,
        BI_RGB
    };
    BITMAPINFO bi = { bh, 0 };
     
    HDC hDCa = CreateCompatibleDC(hDC);
    GetDIBits(hDCa, hBitmap, 0, bm.bmHeight, this->rgba, &bi, DIB_RGB_COLORS);
    DeleteDC(hDCa);
    
    int tmp;
    int i;
    for (i = 0; i < bm.bmWidth * bm.bmHeight; ++i) {
         tmp = this->rgba[i].r;
         this->rgba[i].r = this->rgba[i].b;
         this->rgba[i].b = tmp;
         this->rgba[i].a = 255; // plne
    };
    // prohodit radky
    Aegis::M2::RGBA radek[this->sizex];
    ZeroMemory(radek, sizeof(Aegis::M2::RGBA) * this->sizex);
    for (i = 0; i < sizey>>1; ++i) {
        CopyMemory(radek,&this->rgba[i*this->sizex],this->sizex<<2);
        CopyMemory(&this->rgba[i*this->sizex],&this->rgba[(this->sizey-i-1)*this->sizex],this->sizex<<2);
        CopyMemory(&this->rgba[(this->sizey-i-1)*this->sizex],radek,this->sizex<<2);
    };
    
    return 0;
};

__fastcall int Aegis::F::Bitmap::LoadResource( HDC hDC, LPCTSTR rsrc ) {
    
    return 0;
};


__fastcall int Aegis::F::Bitmap::Draw( HDC hDC, int x, int y, float alpha ) {
    if (alpha > 1.0f) alpha =  1.0f;
    if (alpha <-1.0f) alpha = -1.0f;
    
    COLORREF b; 
    COLORREF bgcol = GetSysColor(COLOR_3DFACE);
    int _r, _g, _b, _a;
    
    if (this->sizex && this->sizey && alpha != 0.0f)
    for (int py = 0; py < this->sizey; ++py)
        for (int px = 0; px < this->sizex; ++px) {
        
        if (alpha > 0.0f)    
            b = ::GetPixel(hDC, x + px, y + py);
        else
            b = bgcol;
          
        _a = (int)(fabsf(alpha) * (float)this->rgba[py*this->sizex + px].a);  
        _r = ((rgba[py*sizex+px].r * _a) >> 8)
             +((GetRValue(b) * (255-_a)) >> 8);
        _g = ((rgba[py*sizex+px].g * _a) >> 8)
             +((GetGValue(b) * (255-_a)) >> 8);
        _b = ((rgba[py*sizex+px].b * _a) >> 8)
             +((GetBValue(b) * (255-_a)) >> 8);

        ::SetPixel(hDC, x+px, y+py, RGB(_r,_g,_b));
    }

    return 0;
};

__fastcall int Aegis::F::Bitmap::Blit( HDC hDC, int x, int y) {
    BITMAPINFO bi = {
        { sizeof(BITMAPINFOHEADER), this->sizex, this->sizey,
          1, 32, BI_RGB }
    };
    COLORREF *bitmap =
    (COLORREF*)GlobalAlloc(GPTR,sizeof(COLORREF)*this->sizex*this->sizey);

    if (this->sizex && this->sizey)
    for (int py = 0;py < this->sizey; ++py)
        for (int px = 0;px < this->sizex; ++px) {
        
        bitmap[(this->sizey-py-1)*this->sizex+px] =
            RGB(rgba[py*this->sizex+px].r, 
                rgba[py*this->sizex+px].g,
                rgba[py*this->sizex+px].b);
    }

    SetDIBitsToDevice(hDC,x,y,this->sizex,this->sizey,0,0,0,this->sizey,
                      bitmap,&bi,DIB_RGB_COLORS);
    GlobalFree(bitmap);

    return 0;
};


__fastcall int Aegis::F::Bitmap::SetPixel( unsigned int x, unsigned int y,
                                           int r, int g, int b, int a ) {
    if (x > this->sizex - 1) return -1;
    if (y > this->sizey - 1) return -1;
    
    register unsigned int offset = y * sizex + x;
    
    this->rgba[ offset ].r = r;
    this->rgba[ offset ].g = g;
    this->rgba[ offset ].b = b;
    this->rgba[ offset ].a = a;
    
    return 0;
};

__fastcall int Aegis::F::Bitmap::SetPixel( unsigned int x, unsigned int y,
                                           DWORD color ) {
    if (x > this->sizex - 1) return -1;
    if (y > this->sizey - 1) return -1;
    
    this->rgba[ y * sizex + x ].dwColor = color;
  
    return 0;
};

__fastcall DWORD Aegis::F::Bitmap::GetPixel(unsigned int x, unsigned int y) const {
    if (x > this->sizex - 1) return this->border.dwColor;
    if (y > this->sizey - 1) return this->border.dwColor;
    
    return this->rgba[ y * sizex + x ].dwColor;
};

__fastcall DWORD Aegis::F::Bitmap::GetPixel(float x, float y, float size) const {
    // TODO: Bilinear Interpolation
    return this->border.dwColor;
};


__fastcall int Aegis::F::Bitmap::GetPixelRed(unsigned int x, unsigned int y) const {
    if (x > this->sizex - 1) return this->border.r;
    if (y > this->sizey - 1) return this->border.r;
    
    return this->rgba[ y * sizex + x ].r;
};

__fastcall int Aegis::F::Bitmap::GetPixelGreen(unsigned int x, unsigned int y)const{
    if (x > this->sizex - 1) return this->border.g;
    if (y > this->sizey - 1) return this->border.g;
    
    return this->rgba[ y * sizex + x ].g;
};

__fastcall int Aegis::F::Bitmap::GetPixelBlue(unsigned int x, unsigned int y) const {
    if (x > this->sizex - 1) return this->border.b;
    if (y > this->sizey - 1) return this->border.b;
    
    return this->rgba[ y * sizex + x ].b;
};

__fastcall int Aegis::F::Bitmap::GetPixelAlpha(unsigned int x, unsigned int y) const {
    if (x > this->sizex - 1) return this->border.a;
    if (y > this->sizey - 1) return this->border.a;
    
    return this->rgba[ y * sizex + x ].a;
};


