/*
  Name: aF_LinMap.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - F (File Formats) - LinMap (aLinMap) base class  
*/

#include "aF_LinMap.hpp"
#include <stdio.h>
#include <zlib.h>
#include <math.h>
#include <new>

static const char *LinMapMagic = "aLinMap";

Aegis::F::LinMap::LinMap() {
    ZeroMemory( this, sizeof(Aegis::F::LinMap) );
    strcpy( this->magic, LinMapMagic );
    
    return;
};    

Aegis::F::LinMap::LinMap( long s ) {
    ZeroMemory( this, sizeof(Aegis::F::LinMap) );
    strcpy( this->magic, LinMapMagic );

    this->size = s;
    this->rgba = new Aegis::M2::RGBA[ this->size ];
    
    return;
};

Aegis::F::LinMap::~LinMap() {
    if (this->rgba) delete[] this->rgba;
    
    return;
};

__fastcall int Aegis::F::LinMap::Load( char *name ) {
    HANDLE hFile = CreateFile( name, GENERIC_READ, FILE_SHARE_READ, NULL,
                               OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    if (hFile == INVALID_HANDLE_VALUE) return GetLastError();
    
    DWORD readBytes = 0;
    unsigned int header_size = sizeof(Aegis::F::LinMap)
                             - sizeof(Aegis::M2::RGBA*);

    Aegis::F::LinMap dumpBitmap;
    if (!ReadFile( hFile, &dumpBitmap, header_size, &readBytes, NULL )) {
        // problem with reading
        CloseHandle( hFile );
        return GetLastError();
    };
    
    // check file structure
    if (strcmpi( dumpBitmap.magic, LinMapMagic )) return -1;
    
    if (readBytes < header_size) return -2;

    // decompress
    int cmperr = 0;
    Aegis::M2::RGBA *buffer = new(std::nothrow)
        Aegis::M2::RGBA[ dumpBitmap.size ];
         
    if (!buffer) return -3; 

    SetFilePointer( hFile, header_size, NULL, FILE_BEGIN );

    switch( dumpBitmap.compression ) {
        case 0: { // no compression
            int readSize = dumpBitmap.size * sizeof(Aegis::M2::RGBA);
            if (!ReadFile( hFile, buffer, readSize, &readBytes, NULL ))
                cmperr = -5;
        } break;

        case 1: { // zlib direct
            int readSize = GetFileSize( hFile, NULL ) - header_size;
            char *temp = new(std::nothrow) char[ readSize ];
            if (!temp) return -3; 
            
            if (!ReadFile( hFile, temp, readSize, &readBytes, NULL ))
                cmperr = -5;
            
            uLongf destLen = dumpBitmap.size * sizeof(Aegis::M2::RGBA);
            int ret = uncompress( (Bytef*) buffer, &destLen,
                                  (const Bytef*) temp, readSize );
            
            if (ret == Z_MEM_ERROR) return -3;
            if (ret == Z_BUF_ERROR) return -6;
            if (ret == Z_DATA_ERROR) return -7;
            if (ret != Z_OK) return -8;
                
            delete temp;
        } break;
        
        // TODO: Add more compressions
     
        
        default:
            cmperr = -9; // unknown compression
    };    
    
    if (cmperr) {
        delete[] buffer;
        CloseHandle( hFile );
        return cmperr;
    };

    // finalize
    this->rgba = buffer;
    
    this->size = dumpBitmap.size;
    this->compression = dumpBitmap.compression;
    this->border = dumpBitmap.border;
    
    this->p0 = dumpBitmap.p0;
    this->p3 = dumpBitmap.p3;
    this->p4 = dumpBitmap.p4;
    
    CloseHandle( hFile );
    return 0;
};
    
__fastcall int Aegis::F::LinMap::Save( char *name, int cmp ) {
    HANDLE hFile = CreateFile( name, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if (hFile == INVALID_HANDLE_VALUE) return GetLastError();

    this->compression = cmp;

    DWORD writtenBytes = 0;
    unsigned int header_size = sizeof(Aegis::F::LinMap)
                             - sizeof(Aegis::M2::RGBA *);

    if (!WriteFile( hFile, this, header_size, &writtenBytes, NULL )) {
        // problem with writting
        CloseHandle( hFile );
        return GetLastError();
    };
    
    void *bufferToWrite;
    unsigned int writeBytes = 0;
    switch( this->compression ) {
        case 0: { // no compression
            bufferToWrite = this->rgba;
            writeBytes = sizeof(Aegis::M2::RGBA) * this->size;
        } break;    
        case 1: { // zlib direct
            writeBytes = compressBound(sizeof(Aegis::M2::RGBA)
                                       * this->size);
            bufferToWrite = GlobalAlloc( GPTR, writeBytes );
            if (!bufferToWrite) {
                CloseHandle( hFile );
                return -3;
            };
            
            int ret = compress2( (Bytef*)bufferToWrite, (uLongf*)&writeBytes,
                                 (Bytef*)this->rgba, sizeof(Aegis::M2::RGBA)
                                 * this->size, 9);

            if (ret != Z_OK) CloseHandle( hFile );
            if (ret == Z_MEM_ERROR) return -3;
            if (ret == Z_BUF_ERROR) return -6;
            if (ret != Z_OK) return -8;
        } break;
        default:
            CloseHandle( hFile );
            return -1;
    };
    
    if (!WriteFile( hFile, bufferToWrite, writeBytes, &writtenBytes, NULL )) {
        // problem with writting
        if (this->compression) GlobalFree(bufferToWrite);
        CloseHandle( hFile );
        
        return GetLastError();
    };
    
    if (this->compression) GlobalFree(bufferToWrite);

    CloseHandle( hFile );
    return 0;
};


__fastcall int Aegis::F::LinMap::Draw( HDC hDC, int x, int y, float alpha ) {
    if (alpha > 1.0f) alpha =  1.0f;
    if (alpha <-1.0f) alpha = -1.0f;
    
    COLORREF b; 
    COLORREF bgcol = GetSysColor(COLOR_3DFACE);
    int _r, _g, _b, _a;
    
    if (this->size && alpha != 0.0f)
    for (int px = 0; px < this->size; ++px) {
        if (alpha > 0.0f)    
            b = ::GetPixel(hDC, x + px, y);
        else
            b = bgcol;
          
        _a = (int)(fabsf(alpha) * (float)this->rgba[px].a);  
        _r = ((rgba[px].r * _a) >> 8)
             +((GetRValue(b) * (255-_a)) >> 8);
        _g = ((rgba[px].g * _a) >> 8)
             +((GetGValue(b) * (255-_a)) >> 8);
        _b = ((rgba[px].b * _a) >> 8)
             +((GetBValue(b) * (255-_a)) >> 8);

        ::SetPixel(hDC, x+px, y, RGB(_r,_g,_b));
    }

    return 0;
};

__fastcall int Aegis::F::LinMap::Blit( HDC hDC, int x, int y) {
    BITMAPINFO bi = {
        { sizeof(BITMAPINFOHEADER), this->size, 1,
          1, 32, BI_RGB }
    };
    COLORREF *bitmap =
    (COLORREF*)GlobalAlloc(GPTR,sizeof(COLORREF)*this->size);

    if (this->size)
    for (int px = 0;px < this->size; ++px)
        bitmap[px] = RGB( rgba[px].r, rgba[px].g, rgba[px].b);

    SetDIBitsToDevice(hDC,x,y,this->size,1,0,0,0,1,bitmap,&bi,DIB_RGB_COLORS);
    GlobalFree(bitmap);

    return 0;
};


__fastcall int Aegis::F::LinMap::SetPixel( unsigned int offset,
                                           int r, int g, int b, int a ) {
    if (offset > this->size - 1) return -1;

    this->rgba[ offset ].r = r;
    this->rgba[ offset ].g = g;
    this->rgba[ offset ].b = b;
    this->rgba[ offset ].a = a;

    return 0;
};    

__fastcall int Aegis::F::LinMap::SetPixel( unsigned int offset, DWORD color ) {
    if (offset > this->size - 1) return -1;

    this->rgba[ offset ].dwColor = color;

    return 0;
};    


__fastcall DWORD Aegis::F::LinMap::GetPixel( unsigned int offset ) const {
    if (offset > this->size - 1) return this->border.dwColor;

    return this->rgba[ offset ].dwColor;
};    

__fastcall int Aegis::F::LinMap::GetPixelRed( unsigned int offset ) const {
    if (offset > this->size - 1) return this->border.r;

    return this->rgba[ offset ].r;
};    

__fastcall int Aegis::F::LinMap::GetPixelGreen( unsigned int offset ) const {
    if (offset > this->size - 1) return this->border.g;

    return this->rgba[ offset ].g;
};    

__fastcall int Aegis::F::LinMap::GetPixelBlue( unsigned int offset ) const {
    if (offset > this->size - 1) return this->border.b;

    return this->rgba[ offset ].b;
};    

__fastcall int Aegis::F::LinMap::GetPixelApha( unsigned int offset ) const {
    if (offset > this->size - 1) return this->border.a;

    return this->rgba[ offset ].a;
};    

__fastcall DWORD Aegis::F::LinMap::GetPixel( float offset, float size ) const {
    // TODO: base on size

    if ((unsigned int)offset > this->size - 1) return this->border.dwColor;
    return this->rgba[ (unsigned int)offset ].dwColor;
};    



