/*
  Name: aF_TexMap.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - F (File Formats) - TexMap (aTexMap) base class  
*/

#include "aF_TexMap.hpp"
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <zlib.h>
#include <math.h>
#include <new>


static const char *TexMapMagic = "aTexMap";

__fastcall Aegis::F::TexMap::TexMap() {
    ZeroMemory( this, sizeof(Aegis::F::TexMap) );
    strcpy( this->magic, TexMapMagic );
    
    return;
};

__fastcall Aegis::F::TexMap::TexMap( Aegis::F::TexMap& tm ) {
    CopyMemory( this, &tm, sizeof(Aegis::F::TexMap) );
    
    int s = 0;
    for (int i = 1; i <= this->size; ++i) s += 1 << i;
    
    this->red   = new(std::nothrow) signed char[s];
    this->green = new(std::nothrow) signed char[s];
    this->blue  = new(std::nothrow) signed char[s];
    this->alpha = new(std::nothrow) signed char[s];
    
    if (this->red && this->green && this->blue && this->alpha ) {
        CopyMemory( this->red,   tm.red,   s * sizeof(signed char) );
        CopyMemory( this->green, tm.green, s * sizeof(signed char) );
        CopyMemory( this->blue,  tm.blue,  s * sizeof(signed char) );
        CopyMemory( this->alpha, tm.alpha, s * sizeof(signed char) );
    };    
    return;
};

__fastcall static Aegis::M2::RGBA
TexMapHelper_Medians( int x1, int y1, int x2, int y2,
                      int mip, Aegis::M2::RGBA *ptr ) {
    Aegis::M2::RGBA pix;

    unsigned char median_largest[4] = { 0, 0, 0, 0 };
    unsigned char median_smallest[4] = { 255, 255, 255, 255 };
    unsigned int sum[4] = { 0, 0, 0, 0 };
    unsigned int n = 0;
        
    for (int x = x1; x < x2; ++x)
        for (int y = y1; y < y2; ++y) {
            int offset = y*(1<<mip) + x;

            ++n;
            for (int i = 0; i < 4; ++i) {
                sum[i] += ptr[offset].color[i];
            
                if (ptr[offset].color[i] > median_largest[i])
                    median_largest[i] = ptr[offset].color[i];
                if (ptr[offset].color[i] < median_smallest[i])
                    median_smallest[i] = ptr[offset].color[i];
            };    
        };    
        
    for (int i = 0; i < 4; ++i) {
        sum[i] /= n;

        if (abs(sum[i] - median_smallest[i]) > 127
         || abs(sum[i] - median_largest[i] ) > 127)
            // use median
            pix.color[i] = (median_largest[i] + median_smallest[i]) >> 1;
        else
            // use average
            pix.color[i] = sum[i];
    };
    
    return pix;
};    

__fastcall static bool
TexMapHelper_Problem( Aegis::F::TexMap &tm,
                      int level, int px, int py, Aegis::M2::RGBA parent,
                      int mip, Aegis::M2::RGBA *ptr ) {
    int num_pixels = 1 << (mip - level);
    int pixel_size = (1 << mip) / num_pixels;
    
    int xoffset = px * pixel_size;
    int yoffset = py * pixel_size;
    
    // get median pixel
    Aegis::M2::RGBA pixel = TexMapHelper_Medians( xoffset, yoffset,
                                                  xoffset + pixel_size,
                                                  yoffset + pixel_size,
                                                  mip, ptr );

    // substract from parent
    Aegis::M2::RGBA final = pixel - parent;
    
    // write to data
    FILE *f = fopen("__w.txt","at");
    fprintf(f,"%d: %3d x %3d\t%04d %04d %04d %04d\n",level,
            xoffset,yoffset,final.red_offset,final.green_offset,
            final.blue_offset,final.alpha_offset);
    fclose(f);
    
//    if (pixel_size > 16) printf("%4d x %4d >>> %4d\t%04d %04d %04d %04d\n",
//    xoffset,yoffset,pixel_size,final.red_offset,final.green_offset,
//    final.blue_offset,final.alpha_offset);

    // do the same for all four sub pixels
    if (level) {
        TexMapHelper_Problem( tm, level-1, 0, 0, final, mip, ptr );
        TexMapHelper_Problem( tm, level-1, 1, 0, final, mip, ptr );
        TexMapHelper_Problem( tm, level-1, 0, 1, final, mip, ptr );
        TexMapHelper_Problem( tm, level-1, 1, 1, final, mip, ptr );
    };    

    return 0;
};

__fastcall Aegis::F::TexMap::TexMap(Aegis::F::Bitmap& bm) {
    ZeroMemory( this, sizeof(Aegis::F::TexMap) );
    strcpy( this->magic, TexMapMagic );
    
    int targetx, targety;
    
    // get closest higher dimension
    for (int i = 0; i < 255; ++i)
        if (((1<<i) >= bm.Width()) && ((1<<i) >= bm.Height())) {
            targetx = (1<<i);
            targety = (1<<i);
            this->size = i;
            break;
        };
    
    // alloc
    int s = 1 << this->size;
    this->red   = new(std::nothrow) signed char[s];
    this->green = new(std::nothrow) signed char[s];
    this->blue  = new(std::nothrow) signed char[s];
    this->alpha = new(std::nothrow) signed char[s];

    if (!this->red || !this->green || !this->blue || !this->alpha) {
        this->size = 0;
        return;
    };    

    // work space
    Aegis::M2::RGBA * ptr = (Aegis::M2::RGBA *)
        GlobalAlloc(GPTR, targetx * targety * sizeof(Aegis::M2::RGBA));
    if (!ptr) return;
                          
    Aegis::M2::RGBA * src = (Aegis::M2::RGBA *) bm.GetPointer();
    
    // create source image
    for (int l = 0; l < bm.Height(); ++l)
        CopyMemory( &ptr[ l * targetx ], &src[ l * bm.Width() ],
                    bm.Width() * sizeof(Aegis::M2::RGBA) );
    
    // get smalest mipmap
    Aegis::M2::RGBA p = TexMapHelper_Medians( 0, 0, targetx, targety,
                                              this->size, ptr );
    // store smallest mipmap
    this->base = p;
    
//    MipMapPointer
   
    if (this->size) {
        TexMapHelper_Problem( *this, this->size-1, 0, 0, p, this->size, ptr );
        TexMapHelper_Problem( *this, this->size-1, 1, 0, p, this->size, ptr );
        TexMapHelper_Problem( *this, this->size-1, 0, 1, p, this->size, ptr );
        TexMapHelper_Problem( *this, this->size-1, 1, 1, p, this->size, ptr );
    };
    
    // clean up
    GlobalFree( ptr );  
    
    return;
};

__fastcall Aegis::F::TexMap::~TexMap() {
    if (this->red)   delete[] this->red;
    if (this->green) delete[] this->green;
    if (this->blue)  delete[] this->blue;
    if (this->alpha) delete[] this->alpha;
    
    return;
};

__fastcall const signed char *
Aegis::F::TexMap::GetPointer( int component, int mipmap ) const {
    if (mipmap >= this->size) return NULL;
    
    unsigned int offset = 0;
    if (mipmap) {
        offset = 1<<(size<<1); 
        
        for (int i=1; i<=mipmap; ++i)
            offset += 1<<((this->size-i)<<1);
    };
        
    switch(component) {
        case 0: return &this->red[ offset ];
        case 1: return &this->green[ offset ];
        case 2: return &this->blue[ offset ];
        case 3: return &this->alpha[ offset ];
    };

    return NULL;    
};//*/

__fastcall int Aegis::F::TexMap::SetPixel( unsigned int x, unsigned int y,
                                           int mipmap, 
                                           Aegis::M2::RGBA clr ) {
    if (mipmap >= this->size) return -1;
    if (x > (1<<(this->size - mipmap))) return -1;
    
    unsigned int offset = 0;
    if (mipmap) {
        offset = 1<<(size<<1); 
        
        for (int i=1; i<=mipmap; ++i)
            offset += 1<<((this->size-i)<<1);
    };
    
    this->red  [ offset ] = clr.red_offset;
    this->green[ offset ] = clr.green_offset;
    this->blue [ offset ] = clr.blue_offset;
    this->alpha[ offset ] = clr.alpha_offset;
    
    return 0;
};    

__fastcall Aegis::F::Bitmap* Aegis::F::TexMap::MakeBitmap() const {
    
    return NULL;
};

__fastcall int Aegis::F::TexMap::Load( const char *name ) {
    
    return 0;
};    

__fastcall int Aegis::F::TexMap::Save( const char *name ) {
    
    return 0;
};    

// GETs - decompose single pixel through all the layers

