/*
  Name: aG1_FogControl.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - G (Graphics) Subsystem - G1 - Fog Control and Cache
*/

#include "aG1_FogControl.hpp"

__fastcall Aegis::G1::FogControlObject::FogControlObject( 
                        const Aegis::G1::Renderer & ren,
                        bool want_emulated_parametric  )
                      : renderer(ren) {
    
    this->constant_fog = false;
    this->parametric_fog = false;
    
    if (want_emulated_parametric)
        this->parametric_capability = false;
    else
        this->parametric_capability = ren.Ext->Is( Aegis::G1::FogCoord );

    this->UpdateEmulateParametricFlag();
    
    // set defaults
    this->glfog_param_set = false;
    this->glfog_enabled = false;
    
    this->glfog_mode = GL_EXP;
    this->glfog_density = 1.0f;
    this->glfog_start = 0.0f;
    this->glfog_end = 1.0f;
    this->glfog_color = Aegis::M2::Color( 0.0f, 0.0f, 0.0f, 0.0f );

//    this->mode_stack.reserve(6144);
//    this->fparam_stack.reserve(6144);
//    this->color_stack.reserve(2048);
    
    return;
};    

__fastcall void Aegis::G1::FogControlObject::UpdateEmulateParametricFlag() {
    if (!this->parametric_capability
        || (this->parametric_fog && this->constant_fog))
        this->emulate_parametric = true;
    else
        this->emulate_parametric = false;
    return;
};    

__fastcall void Aegis::G1::FogControlObject::Enable( bool en ) {
    if (this->constant_fog == en) return;

    this->constant_fog = en;
    this->UpdateEmulateParametricFlag();
    this->UpdateFogMode();
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::ParametricEnable( bool en ) {
    if (this->parametric_fog == en) return;
    
    this->parametric_fog = en;
    this->UpdateEmulateParametricFlag();
    this->UpdateFogMode();
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::UpdateFogMode() {
    if (!this->parametric_capability) {
        // no need to switch parametric/global
        if (this->constant_fog) {
            if (!this->glfog_enabled) {
                this->glfog_enabled = true;
                glEnable( GL_FOG );
            };
        } else {
            if (this->glfog_enabled) {
                this->glfog_enabled = false;
                glDisable( GL_FOG );
            };    
        };
    } else {
        if (this->constant_fog || this->parametric_fog) {
            if (!this->glfog_enabled) {
                this->glfog_enabled = true;
                glEnable( GL_FOG );
            };
        } else {
            if (this->glfog_enabled) {
                this->glfog_enabled = false;
                glDisable( GL_FOG );
            };    
        };

        // set mode, prefer global constant fog
        if (this->parametric_fog && !this->constant_fog) {
            if (!this->glfog_param_set) {
                this->glfog_param_set = true;
                glFogi( GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT );
            };
        } else {
            if (this->glfog_param_set) {
                this->glfog_param_set = false;
                glFogi( GL_FOG_COORDINATE_SOURCE_EXT, GL_FRAGMENT_DEPTH_EXT );
            };
        };
        
    };
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::SetMode( int mode ) {
    if (this->glfog_mode == mode) return;
    if (mode != GL_LINEAR && mode != GL_EXP && mode != GL_EXP2) return;
    
    this->glfog_mode = mode;
    glFogi( GL_FOG_MODE, mode );
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::SetDensity( float density ) {
    if (this->glfog_density == density) return;
    
    this->glfog_density = density;
    glFogf( GL_FOG_DENSITY, density );
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::SetLinearStart( float start ) {
    if (this->glfog_start == start) return;
    
    this->glfog_start = start;
    glFogf( GL_FOG_START, start );
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::SetLinearEnd( float end ) {
    if (this->glfog_end == end) return;
    
    this->glfog_end = end;
    glFogf( GL_FOG_END, end );
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::SetColor( Aegis::M2::Color clr ) {
    if (this->glfog_color == clr) return;
    
    this->glfog_color = clr;
    glFogfv( GL_FOG_COLOR, clr.c );
    
    return;
};

__fastcall void Aegis::G1::FogControlObject::Push() {
    this->mode_stack.push( this->parametric_fog );
    this->mode_stack.push( this->constant_fog );
    this->mode_stack.push( this->glfog_mode );
    
    this->fparam_stack.push( this->glfog_end );
    this->fparam_stack.push( this->glfog_start );
    this->fparam_stack.push( this->glfog_density );
    
    this->color_stack.push( this->glfog_color );
    
    return;
};    

__fastcall void Aegis::G1::FogControlObject::Pop() {
    this->SetDensity(     this->fparam_stack.top() ); this->fparam_stack.pop();
    this->SetLinearStart( this->fparam_stack.top() ); this->fparam_stack.pop();
    this->SetLinearEnd(   this->fparam_stack.top() ); this->fparam_stack.pop();
    
    this->SetColor(        this->color_stack.top() ); this->color_stack.pop();
    
    this->SetMode(          this->mode_stack.top() ); this->mode_stack.pop();
    this->Enable(           this->mode_stack.top() ); this->mode_stack.pop();
    this->ParametricEnable( this->mode_stack.top() ); this->mode_stack.pop();
    
    return;
};    


                
