/*
  Name: aG1_LightControl.HPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - G (Graphics) Subsystem - G1 - Light Control
*/

#include "aG1_LightControl.hpp"

__fastcall Aegis::G1::LightControlObject::
           LightControlObject( const Aegis::G1::Renderer &ren )
           : renderer(ren) {
    
    glGetIntegerv( GL_MAX_LIGHTS, &this->n_lights );
    
    this->lighting = false;
    this->light = new LightState[ this->n_lights ];
    
    for (unsigned int i = 0; i < this->n_lights; ++i) {
        this->light[i].enabled = false;
        this->light[i].positional = false;

        this->light[i].ambient = Aegis::M2::Color( 0.0f, 0.0f, 0.0f, 1.0f );
        this->light[i].diffuse = Aegis::M2::Color( 1.0f, 1.0f, 1.0f, 1.0f );
        this->light[i].specular = Aegis::M2::Color( 1.0f, 1.0f, 1.0f, 1.0f );

        this->light[i].direction = Aegis::M2::Normal( 0.0f, 0.0f, 1.0f );

        this->light[i].exponent = 0.0f;
        this->light[i].cutoff = 180.0f;

        this->light[i].attenuation_quadratic = 0.0f;
        this->light[i].attenuation_linear = 0.0f;
        this->light[i].attenuation_constant = 1.0f;
    };
    
    this->local_viewer = false;
    this->double_sided = false;
    this->global_ambient = Aegis::M2::Color( 0.2f, 0.2f, 0.2f, 1.0f );
    
    return;
};

__fastcall Aegis::G1::LightControlObject::~LightControlObject() {
    delete[] this->light;
};

__fastcall void Aegis::G1::LightControlObject::Enable( bool en ) {
    if (en == this->lighting) return;
    this->lighting = en;

    if (en) {
        glDisable( GL_COLOR_MATERIAL );
        glEnable( GL_LIGHTING );
    } else {
        glEnable( GL_COLOR_MATERIAL );
        glDisable( GL_LIGHTING );
    };
    
    return;
};

__fastcall void Aegis::G1::LightControlObject::EnableLight( int l, bool en ) {
    if (l < 0 || l >= this->n_lights) return;

    if (this->light[l].enabled == en) return;
    this->light[l].enabled = en;

    if (en) {
        glEnable( GL_LIGHT0 + l );
    } else {
        glDisable( GL_LIGHT0 + l );
    };
    
    return;
};

_fastcall void Aegis::G1::LightControlObject::DisableAll() {
    for (int i = 0; i < this->n_lights; ++i) this->EnableLight( i, false );
    return;
};

__fastcall void Aegis::G1::LightControlObject::LocalViewer( bool lv ) {
    if (lv == this->local_viewer) return;
    this->local_viewer = lv;
    
    glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, lv );
    
    return;
};

__fastcall void Aegis::G1::LightControlObject::DoubleSided( bool ds ) {
    if (ds == this->double_sided) return;
    this->double_sided = ds;
    
    glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, ds );
    
    return;
};

__fastcall void
Aegis::G1::LightControlObject::SetGlobalAmbient( const Aegis::M2::Color & amb ) {
    if (amb == this->global_ambient) return;
    this->global_ambient = amb;
    
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, amb.c );
    
    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetPositional( int l, Aegis::M2::Vertex pos ) {
    if (l < 0 || l >= this->n_lights) return;
    
    pos.AdjustWeight();

    this->light[l].positional = true;
    this->light[l].position = pos;
    
    pos.w = 1.0; // because of this it cannot be reference
    glLightfv( GL_LIGHT0 + l, GL_POSITION, pos.v );

    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetDirectional( int l, Aegis::M2::Normal dir ) {
    if (l < 0 || l >= this->n_lights) return;
    
    this->light[l].positional = false;
    this->light[l].direction = dir;
    
    dir.v[3] = 0.0; // because of this it cannot be reference
    glLightfv( GL_LIGHT0 + l, GL_POSITION, dir.v );

    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetAmbient(int l, const Aegis::M2::Color& c) {
    if (l < 0 || l >= this->n_lights) return;
    if (this->light[l].ambient == c) return;
    
    this->light[l].ambient = c;
    glLightfv( GL_LIGHT0 + l, GL_AMBIENT, c.c );
    
    return;
};
    

__fastcall void 
Aegis::G1::LightControlObject::SetDiffuse(int l, const Aegis::M2::Color& c) {
    if (l < 0 || l >= this->n_lights) return;
    if (this->light[l].diffuse == c) return;
    
    this->light[l].diffuse = c;
    glLightfv( GL_LIGHT0 + l, GL_DIFFUSE, c.c );
    
    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetSpecular(int l, const Aegis::M2::Color& c) {
    if (l < 0 || l >= this->n_lights) return;
    if (this->light[l].specular == c) return;
    
    this->light[l].specular = c;
    glLightfv( GL_LIGHT0 + l, GL_SPECULAR, c.c );
    
    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetSpotDirection( int l,
                                                 const Aegis::M2::Normal & dir) {
    if (l < 0 || l >= this->n_lights) return;
    if (this->light[l].spot_direction == dir) return;
    this->light[l].spot_direction = dir;
    
    glLightfv( GL_LIGHT0 + l, GL_SPOT_DIRECTION, dir.v );
    
    return;
};

__fastcall void
Aegis::G1::LightControlObject::SetSpotExponent( int l, float exponent ) {
    if (l < 0 || l >= this->n_lights) return;
    if (exponent < 0.0f || exponent > 128.0f ) return;

    if (this->light[l].exponent == exponent) return;
    this->light[l].exponent = exponent;
    
    glLightf( GL_LIGHT0 + l, GL_SPOT_EXPONENT, exponent );
    
    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetSpotCutOff( int l, float cutoff ) {
    if (l < 0 || l >= this->n_lights) return;
    if ((cutoff < 0.0f || cutoff > 90.0f) && cutoff != 180.0f) return;

    if (this->light[l].cutoff == cutoff) return;
    this->light[l].cutoff = cutoff;
    
    glLightf( GL_LIGHT0 + l, GL_SPOT_CUTOFF, cutoff );
        
    return;
};

__fastcall void 
Aegis::G1::LightControlObject::SetAttenuation( int light, float constant,
                                               float linear, float quad ) {
    if (light < 0 || light >= this->n_lights) return;
    
    if (this->light[light].attenuation_constant != constant) {
        this->light[light].attenuation_constant = constant;
        
        glLightf( GL_LIGHT0 + light, GL_CONSTANT_ATTENUATION , constant );
    };
    if (this->light[light].attenuation_linear != linear) {
        this->light[light].attenuation_linear = linear;
        
        glLightf( GL_LIGHT0 + light, GL_LINEAR_ATTENUATION, linear );
    };
    if (this->light[light].attenuation_quadratic != quad) {
        this->light[light].attenuation_quadratic = quad;
        
        glLightf( GL_LIGHT0 + light, GL_QUADRATIC_ATTENUATION, quad );
    };
    
    return;
};



