/*
  Name: aG2_CubeMap.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - G (Graphics) Subsystem - G2 - CubeMap Class
*/

#include "aG2_CubeMap.hpp"
#include "../../M/M2/aM2_Normal.hpp"

__fastcall int Aegis::G2::CubeMap::Build( const Aegis::G1::Renderer & renderer,
                                          const Aegis::F::Bitmap bm[6] ,
                                          bool mipmap, int filter,
                                          float aniso ) {
    renderer.Textures->Push();
    renderer.Textures->Enable(0,GL_TEXTURE_CUBE_MAP_ARB);

    this->SetParams( renderer, mipmap, filter, aniso ); 
    this->SetPriority( renderer, Aegis::G2::TexturePriorityNormal );
    
    // create texture
    if (mipmap && !renderer.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
        // generate textures
        for (int i = 0; i < 6; ++i) 
            gluBuild2DMipmaps( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + i, GL_RGBA8,
                               bm[i].Width(), bm[i].Height(),
                               GL_RGBA, GL_UNSIGNED_BYTE, bm[i].GetPointer() );
    else {
        int maxsize;
        glGetIntegerv( GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB, &maxsize );
        
        // all six textures
        for (int tex = 0; tex < 6; ++tex )
        // consider revision of this IF, this really sucks
        if (!((!(bm[tex].Width()&bm[tex].Width()-1))
              &&
              (!(bm[tex].Height()&bm[tex].Height()-1)))
            || bm[tex].Width() > maxsize || bm[tex].Height() > maxsize
            || bm[tex].Width() != bm[tex].Height() ) {
                
            // need to resize
            int targetx, targety, limit = 0;
        
            // get maximum texture shift
            do { maxsize >>= 1; ++limit; } while(maxsize);    
            
            // get closest higher dimension
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm[tex].Width())) {
                targetx=(1<<i);
                break;
            }
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm[tex].Height())){
                targety=(1<<i);
                break;
            }
            
            if (targetx != targety) { // ensure quad
                targetx = targety = (targetx >? targety);
            };
            
            Aegis::M2::RGBA *ptr = (Aegis::M2::RGBA *)
                GlobalAlloc(GPTR, targetx*targety*sizeof(Aegis::M2::RGBA));

            gluScaleImage( GL_RGBA, bm[tex].Width(),bm[tex].Height(),
                           GL_UNSIGNED_BYTE, bm[tex].GetPointer(),
                           targetx, targety, GL_UNSIGNED_BYTE, ptr );
            
            // generate texture
            glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + tex,
                          0, GL_RGBA8, targetx, targety,
                          0, GL_RGBA, GL_UNSIGNED_BYTE, ptr );
                          
            // clean up
            GlobalFree( ptr );  
        } else { // use bitmap data directly
            // generate texture
            glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + tex,
                          0, GL_RGBA8, bm[tex].Width(), bm[tex].Height(),
                          0, GL_RGBA, GL_UNSIGNED_BYTE, bm[tex].GetPointer() );
        };    
    };

    renderer.Textures->Pop();
    return 0;
};

// Generate normalisation cube map
// Original code from: www.paulsprojects.net
// This code is modified for Aegis Engine
int Aegis::G2::CubeMap::BuildNormalizationMap(const Aegis::G1::Renderer & ren) {
    // prepare map
    ren.Textures->Push();
    ren.Textures->Enable(0,GL_TEXTURE_CUBE_MAP_ARB);

    this->SetParams( ren, 0 ); // no mipmaps !!!
    this->SetPriority( ren, Aegis::G2::TexturePriorityAboveNormal );
    
    // preapare for creating bitmaps
    unsigned char data[4096]; // 32*32 * 4
    memset( data, 255, 4096 );
    
	const int size = 32;
	float offset = 0.5f;
	float halfSize = 16.0f;
	Aegis::M2::Normal tempVector;
	unsigned char *ptr;
    
    int i, j;
    
    // GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB
    ptr = data;
    
    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct(  halfSize,
                                  -(j+offset-halfSize),
                                  -(i+offset-halfSize) );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);


	// GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB
    ptr = data;

    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct( -halfSize,
                                  -(j+offset-halfSize),
                                   (i+offset-halfSize) );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    // GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB
    ptr = data;

    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct(  (i+offset-halfSize),
                                   halfSize,
                                   (j+offset-halfSize) );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    // GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB
    ptr = data;

    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct(  (i+offset-halfSize),
                                  -halfSize,
                                  -(j+offset-halfSize) );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    // GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB
    ptr = data;

    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct(  (i+offset-halfSize),
                                  -(j+offset-halfSize),
                                   halfSize );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    // GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
    ptr = data;

    for(j = 0; j < size; ++j)
        for(i = 0; i < size; ++i) {
            tempVector.Construct( -(i+offset-halfSize),
                                  -(j+offset-halfSize),
                                   halfSize );
			tempVector.Pack01();

			ptr[0] = (unsigned char)(tempVector.x * 255.0f);
			ptr[1] = (unsigned char)(tempVector.y * 255.0f);
			ptr[2] = (unsigned char)(tempVector.z * 255.0f);
			ptr += 4;
		}
	glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB,
                  0, GL_RGBA8, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    ren.Textures->Pop();

    return 0;
};

__fastcall Aegis::F::Bitmap *
Aegis::G2::CubeMap::MakeBitmap( const Aegis::G1::Renderer & rn,
                                int side, int mipmap ) const {
    
    rn.Textures->Push();
    rn.Textures->Enable(0,GL_TEXTURE_CUBE_MAP_ARB);
    int tex, w, h;

    // yes, we could rely here on tex = POS_X + side, but ...
    switch(side) {
        default:
        case 0: tex = GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB; break;
        case 1: tex = GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB; break;
        case 2: tex = GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB; break;
        case 3: tex = GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB; break;
        case 4: tex = GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB; break;
        case 5: tex = GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB; break;
    };    
    
    this->Bind( rn );
    glGetTexLevelParameteriv( tex, mipmap, GL_TEXTURE_WIDTH, &w );
    glGetTexLevelParameteriv( tex, mipmap, GL_TEXTURE_HEIGHT, &h );

    Aegis::F::Bitmap *bitmap = new Aegis::F::Bitmap( w, h );
    if (!bitmap->Pointer()) { // not enough memory
        delete bitmap;
        return NULL;
    };    
    
    glGetTexImage( tex, mipmap, GL_RGBA, GL_UNSIGNED_BYTE,
                   bitmap->Pointer() );
    rn.Textures->Pop();
    
    return bitmap;
};


__fastcall void Aegis::G2::CubeMap::SetParams( const Aegis::G1::Renderer &rn,
                                               bool mipmap, int filter,
                                               float aniso ) {
    rn.Textures->Enable(0,GL_TEXTURE_CUBE_MAP_ARB);
    this->Bind( rn );

    // set filter
    if (mipmap) {
//      TODO: Get more info if CubeMaps are really unsupported
//        if (rn.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
//            glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE );

        if (filter < 0) filter = 0;
        if (filter > 2) filter = 2;
        switch(filter) {
            case 0:
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAG_FILTER, GL_NEAREST );
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER,
                                 GL_NEAREST_MIPMAP_NEAREST );
            break;
            case 1:
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, 
                                 GL_LINEAR_MIPMAP_NEAREST );
            break;
            case 2:
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, 
                                 GL_LINEAR_MIPMAP_LINEAR );
            break;
        };
    } else {
        if (filter < 0) filter = 0;
        if (filter > 1) filter = 1;
        switch(filter) {
            case 0:
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAG_FILTER, GL_NEAREST );
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MIN_FILTER, GL_NEAREST );
            break;
            case 1:
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MIN_FILTER, GL_LINEAR );
            break;
        };    
    };
    
    // TODO: Consider, do we really need anisotropic filtering
    //       for cubemaps ??? Sounds like nonsense.
    if (aniso > 1.0f || aniso < 0.0f)
        if (rn.Ext->Is( Aegis::G1::TextureAnisotropicFilter )) {
            float max_aniso;
            glGetFloatv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &max_aniso );
            
            if (aniso < 0.0f)
                glTexParameterf( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAX_ANISOTROPY_EXT, max_aniso );
            else
                glTexParameterf( GL_TEXTURE_CUBE_MAP_ARB,
                                 GL_TEXTURE_MAX_ANISOTROPY_EXT,
                                 max_aniso <? aniso );
        };
            
    return;
};    

__fastcall void Aegis::G2::CubeMap::SetPriority( const Aegis::G1::Renderer& rn, 
                                                 float priority ) {
    rn.Textures->Enable(0,GL_TEXTURE_CUBE_MAP_ARB);
    this->Bind( rn );
    glTexParameterf( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_PRIORITY, priority );

    return;
};    


