/*
  Name: aG2_Element.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - G (Graphics) Subsystem - G2 - Element Class
*/

#include "aG2_Element.hpp"
#include <typeinfo>

Aegis::G2::Element::Element(const Aegis::G1::Renderer & ren) : renderer(ren) {
    if (this->renderer.Ext->Is(Aegis::G1::VertexBufferObject)) {
        // VBO
        this->renderer.Ext->glGenBuffersARB( 1, &this->object );
        
        this->vbo_vertex_offset = 0;
        this->vbo_normal_offset = 0;
        this->vbo_color_offset = 0;
        this->vbo_fog_offset = 0;
        
        for (int i = 0; i < AEGIS_G1_MULTITEXTURE; ++i)
            this->vbo_coord_offset[i] = 0;
        
    } else {
        // DISPLAY LIST
        //  - will be damned slow if modified each frame
        this->object = glGenLists(1);
    };
    
    this->Reset();
    
    return;
};

Aegis::G2::Element::~Element() {
    if (this->renderer.Ext->Is(Aegis::G1::VertexBufferObject)) {
        this->renderer.Ext->glDeleteBuffersARB( 1, &this->object );
    } else {
        glDeleteLists( this->object, 1 );
    };
    
    return;
};

__fastcall void Aegis::G2::Element::Reset() {
    this->uncached = true;
    
    this->pure_faces.clear();
    this->multifaces.clear();

    // clear otherfaces
    for (unsigned int i = 0; i < this->otherfaces.size(); ++i) {
        Aegis::G2::Face * pface = this->otherfaces[i];
        delete pface;
    };
    this->otherfaces.clear();

    // following is just for sure    
    this->vbo_modify_texture_points.clear();
    this->vbo_modify_multitexture_points.clear();
    
    this->vbo_vertex_offset = 0;
    this->vbo_normal_offset = 0;
    this->vbo_color_offset = 0;
    this->vbo_fog_offset = 0;
    
    for (int i = 0; i < AEGIS_G1_MULTITEXTURE; ++i)
        this->vbo_coord_offset[i] = 0;
    
    return;
};

__fastcall void Aegis::G2::Element::Draw() {
    if (this->uncached) {
        this->uncached = false;
        
        // recreate the data
        if (this->renderer.Ext->Is(Aegis::G1::VertexBufferObject)) {
            this->RebuildVBO();
        } else {
            this->RebuildDL();
        };
    };
    
    if (!this->pure_faces.size() &&
        !this->multifaces.size() &&
        !this->otherfaces.size()) return; // no need to render
    
    // Save current states that might be modified by Element
    this->renderer.Textures->Push();
    this->renderer.Textures->DisableAll();
    this->renderer.Blend->Push();
    
    // Invoke data
    if (this->renderer.Ext->Is(Aegis::G1::VertexBufferObject)) {
        this->renderer.Ext->glBindBufferARB( GL_ARRAY_BUFFER_ARB, this->object );

        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        glEnableClientState(GL_COLOR_ARRAY);

        this->renderer.Textures->Current(0);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        if (this->renderer.Ext->Is(Aegis::G1::FogCoord))
            glEnableClientState(GL_FOG_COORDINATE_ARRAY_EXT);
        
        // set offsets
        glVertexPointer(4, GL_FLOAT, 0, (void*)this->vbo_vertex_offset );
        glNormalPointer(GL_FLOAT, 0, (void*)this->vbo_normal_offset );
        glColorPointer(4, GL_FLOAT, 0, (void*)this->vbo_color_offset );
        glTexCoordPointer(4, GL_FLOAT, 0, (void*)this->vbo_coord_offset[0] );

        if (this->renderer.Ext->Is(Aegis::G1::FogCoord))
            this->renderer.Ext->glFogCoordPointerEXT(GL_FLOAT, 0,
                                                     (void*)this->vbo_fog_offset );

        // Single faces
        if (this->pure_faces.size())
            if (this->vbo_modify_texture_points.size() > 1) {
                // render parts and change the textures
            
            } else {
                // set the first and only texture
                this->pure_faces[0].texunit.Bind( this->renderer );
                // the simplest way, just render the buffer
                glDrawArrays(GL_TRIANGLES, 0, 3 * this->pure_faces.size());
            };
        
        // TODO: MultiFaces invoke here
        if (this->multifaces.size())
            if (this->vbo_modify_multitexture_points.size() > 1) {
                // render parts and change the textures
            
            } else {
                // set the first set of  textures
//                for (int i = 0; i <
//                this->multifaces[0].texunit[i].Bind( this->renderer, i );

                // the simplest way, just render the buffer
            // TODO: Continue with rendering of MultiFaces
//                if (this->multifaces.size())
//                    glDrawArrays(GL_TRIANGLES, 3 * this->pure_faces.size(),
//                                               3 * this->multifaces.size() );
            };
                

        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_COLOR_ARRAY);

        this->renderer.Textures->Current(0);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        if (this->renderer.Ext->Is(Aegis::G1::FogCoord))
            glDisableClientState(GL_FOG_COORDINATE_ARRAY_EXT);
        
        
    } else {
        // Display list rendering is the easiest
        glPushAttrib( GL_ALL_ATTRIB_BITS ); // unfortunately required
        glCallList( this->object );
        glPopAttrib();
    };

    // Finalize with per-face rendering of unknown faces
    // TODO: FIXME: Problem with this, see Add function
//    for (int i = 0; i < this->otherfaces.size(); ++i)
//        this->otherfaces[i].Draw( this->renderer );
    
    this->renderer.Blend->Pop();
    this->renderer.Textures->Pop();

    return;
};

__fastcall void Aegis::G2::Element::RebuildVBO() {
    // IMPORTANT: Convert colors from FLOAT to UNSIGNED CHAR !!!!!!!!!!!!!1
    
    
    return;
};

__fastcall void Aegis::G2::Element::RebuildDL() {
    
    return;
};

__fastcall bool Aegis::G2::Element::Add(const Aegis::G2::Face & f) {
    // check the face type
    if (typeid(f) == typeid(Aegis::G2::Face)) {
        // single face
        
        
    } else {
        // unknown face, uncached
        // TODO: How to solve this ???
        //       We do not know the size, so we cannot allocate copy
        //       of the Face.
    };
    
    return false;
};







