/*
  Name: aG2_Texture.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - G (Graphics) Subsystem - G2 - Texture Class
*/

#include "aG2_Texture.hpp"
#include <gl/glu.h>
#include <stdio.h>

namespace Aegis {
    namespace G2 {
        const float TexturePriorityLowest = 0.0f;
        const float TexturePriorityLow = 0.1f;
        const float TexturePriorityBelowNormal = 0.3f;
        const float TexturePriorityNormal = 0.4f;
        const float TexturePriorityAboveNormal = 0.5f;
        const float TexturePriorityHigh = 0.7f;
        const float TexturePriorityCritical = 0.9f;
        const float TexturePriorityHighest = 1.0f;
        
    };
};

Aegis::G2::Texture::Texture() {
    glGenTextures( 1, &this->texture );
    return;
};    

Aegis::G2::Texture::~Texture() {
    glDeleteTextures( 1, &this->texture );
    return;
};    

__fastcall void Aegis::G2::Texture::SetParams( const Aegis::G1::Renderer &rn,
                                               bool mipmap, int filter,
                                               float aniso, bool repeat ) {
    rn.Textures->Enable(0,GL_TEXTURE_2D);
    this->Bind( rn );

    // set filter
    if (mipmap) {
        // possible error when unsupported can be ignored
        if (rn.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
            glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE );

        if (filter < 0) filter = 0;
        if (filter > 2) filter = 2;
        switch(filter) {
            case 0:
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                 GL_NEAREST_MIPMAP_NEAREST );
            break;
            case 1:
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
                                 GL_LINEAR_MIPMAP_NEAREST );
            break;
            case 2:
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
                                 GL_LINEAR_MIPMAP_LINEAR );
            break;
        };
    } else {
        if (filter < 0) filter = 0;
        if (filter > 1) filter = 1;
        switch(filter) {
            case 0:
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            break;
            case 1:
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
                glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
            break;
        };    
    };
    
    if (aniso > 1.0f || aniso < 0.0f)
        if (rn.Ext->Is( Aegis::G1::TextureAnisotropicFilter )) {
            float max_aniso;
            glGetFloatv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &max_aniso );
            
            if (aniso < 0.0f)
                glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
                                 max_aniso );
            else
                glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
                                 max_aniso <? aniso );
        };
            
    
    // set repeating
    if (repeat) {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
    } else {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
    };

    return;
};    

__fastcall void Aegis::G2::Texture::SetPriority( const Aegis::G1::Renderer& rn, 
                                                 float priority ) {
    rn.Textures->Enable(0,GL_TEXTURE_2D);
    this->Bind( rn );

    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_PRIORITY, priority );

    return;
};    

__fastcall int Aegis::G2::Texture::Build( const Aegis::G1::Renderer& renderer, 
                                          const Aegis::F::Bitmap &bm, bool mipmap,
                                          int filter, float aniso, bool repeat ){
    renderer.Textures->Push();
    renderer.Textures->Enable(0,GL_TEXTURE_2D);

    // set filter (bind is inside)
    this->SetParams( renderer, mipmap, filter, aniso, repeat );
    this->SetPriority( renderer, Aegis::G2::TexturePriorityNormal );
    
//    printf("Build: %d\n",glIsEnabled(GL_TEXTURE_2D));

    // create texture
    if (mipmap && !renderer.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
        // generate texture
        gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGBA8, bm.Width(), bm.Height(),
                           GL_RGBA, GL_UNSIGNED_BYTE, bm.GetPointer() );
    else {
        int maxsize;
        glGetIntegerv( GL_MAX_TEXTURE_SIZE, &maxsize );
        
        if (!((!(bm.Width()&bm.Width()-1)) && (!(bm.Height()&bm.Height()-1)))
            || bm.Width() > maxsize || bm.Height() > maxsize ) {
                
            // need to resize
            int targetx, targety, limit = 0;
        
            // get maximum texture shift
            do { maxsize >>= 1; ++limit; } while(maxsize);    
            
            // get closest higher dimension
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Width())) {
                targetx=(1<<i);
                break;
            }
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Height())){
                targety=(1<<i);
                break;
            }
            
            Aegis::M2::RGBA *ptr = (Aegis::M2::RGBA *)
                GlobalAlloc(GPTR, targetx*targety*sizeof(Aegis::M2::RGBA));

            gluScaleImage( GL_RGBA, bm.Width(),bm.Height(), GL_UNSIGNED_BYTE,
                           bm.GetPointer(), targetx,targety, GL_UNSIGNED_BYTE,
                           ptr );
            
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, targetx, targety,
                          0, GL_RGBA, GL_UNSIGNED_BYTE, ptr );
                          
            // clean up
            GlobalFree( ptr );  
        } else { // use bitmap data directly
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, bm.Width(), bm.Height(),
                          0, GL_RGBA, GL_UNSIGNED_BYTE, bm.GetPointer() );
        };    
    };    
    
    renderer.Textures->Pop();

    return 0;
};    

__fastcall int
Aegis::G2::Texture::BuildAlphaMap( const Aegis::G1::Renderer& renderer, 
                                   const Aegis::F::Bitmap &bm, bool mipmap,
                                   int filter, float aniso, bool repeat ){
    renderer.Textures->Push();
    renderer.Textures->Enable(0,GL_TEXTURE_2D);

    // set filter (bind is inside)
    this->SetParams( renderer, mipmap, filter, aniso, repeat );
    this->SetPriority( renderer, Aegis::G2::TexturePriorityNormal );
    
    // prepare
    unsigned char * alphamap = new unsigned char[ bm.Width() * bm.Height() ];
    for (int y = 0; y < bm.Height(); ++y)
        for (int x = 0; x < bm.Width(); ++x) {
            alphamap[ y*bm.Width() + x ] = bm.GetPixelAlpha(x,y);
        };
    
    // create texture
    if (mipmap && !renderer.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
        // generate texture
        gluBuild2DMipmaps( GL_TEXTURE_2D, 1/*GL_ALPHA8*/, bm.Width(), bm.Height(),
                           GL_ALPHA, GL_UNSIGNED_BYTE, alphamap );
    else {
        int maxsize;
        glGetIntegerv( GL_MAX_TEXTURE_SIZE, &maxsize );
        
        if (!((!(bm.Width()&bm.Width()-1)) && (!(bm.Height()&bm.Height()-1)))
            || bm.Width() > maxsize || bm.Height() > maxsize ) {
                
            // need to resize
            int targetx, targety, limit = 0;
        
            // get maximum texture shift
            do { maxsize >>= 1; ++limit; } while(maxsize);    
            
            // get closest higher dimension
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Width())) {
                targetx=(1<<i);
                break;
            }
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Height())){
                targety=(1<<i);
                break;
            }
            
            Aegis::M2::RGBA *ptr = (Aegis::M2::RGBA *)
                GlobalAlloc(GPTR, targetx*targety*sizeof(Aegis::M2::RGBA));

            gluScaleImage( GL_ALPHA, bm.Width(),bm.Height(), GL_UNSIGNED_BYTE,
                           alphamap, targetx,targety, GL_UNSIGNED_BYTE,
                           ptr );
            
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA8, targetx, targety,
                          0, GL_ALPHA, GL_UNSIGNED_BYTE, alphamap );
                          
            // clean up
            GlobalFree( ptr );  
        } else { // use bitmap data directly
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA8, bm.Width(), bm.Height(),
                          0, GL_ALPHA, GL_UNSIGNED_BYTE, alphamap );
        };    
    };    
    
    renderer.Textures->Pop();

    return 0;
};    


__fastcall int
Aegis::G2::Texture::BuildLuminanceAlphaMap( const Aegis::G1::Renderer& renderer, 
                                            const Aegis::F::Bitmap &bm, bool mipmap,
                                            int filter, float aniso, bool repeat ){
    renderer.Textures->Push();
    renderer.Textures->Enable(0,GL_TEXTURE_2D);

    // set filter (bind is inside)
    this->SetParams( renderer, mipmap, filter, aniso, repeat );
    this->SetPriority( renderer, Aegis::G2::TexturePriorityNormal );
    
    // prepare
    unsigned char * map = new unsigned char[ 2 * bm.Width() * bm.Height() ];
    for (int y = 0; y < bm.Height(); ++y)
        for (int x = 0; x < bm.Width(); ++x) {
            int c = bm.GetPixelRed(x,y)
                  + bm.GetPixelGreen(x,y)
                  + bm.GetPixelBlue(x,y);
            map[ ((y*bm.Width() + x)<<1) + 0 ] = c / 3;
            map[ ((y*bm.Width() + x)<<1) + 1 ] = bm.GetPixelAlpha(x,y);
        };
    
    // create texture
    if (mipmap && !renderer.Ext->Is( Aegis::G1::TextureGenerateMipmap ))
        // generate texture
        gluBuild2DMipmaps( GL_TEXTURE_2D, 2/*GL_LUMINANCE8_ALPHA8*/,
                            bm.Width(), bm.Height(),
                           GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, map );
    else {
        int maxsize;
        glGetIntegerv( GL_MAX_TEXTURE_SIZE, &maxsize );
        
        if (!((!(bm.Width()&bm.Width()-1)) && (!(bm.Height()&bm.Height()-1)))
            || bm.Width() > maxsize || bm.Height() > maxsize ) {
                
            // need to resize
            int targetx, targety, limit = 0;
        
            // get maximum texture shift
            do { maxsize >>= 1; ++limit; } while(maxsize);    
            
            // get closest higher dimension
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Width())) {
                targetx=(1<<i);
                break;
            }
            for (int i=1;i<limit;++i) if ((1<<i)>=(bm.Height())){
                targety=(1<<i);
                break;
            }
            
            Aegis::M2::RGBA *ptr = (Aegis::M2::RGBA *)
                GlobalAlloc(GPTR, targetx*targety*sizeof(Aegis::M2::RGBA));

            gluScaleImage( GL_LUMINANCE_ALPHA, bm.Width(),bm.Height(),
                           GL_UNSIGNED_BYTE,
                           map, targetx,targety, GL_UNSIGNED_BYTE,
                           ptr );
            
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE8_ALPHA8,
                          targetx, targety,
                          0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, map );
                          
            // clean up
            GlobalFree( ptr );  
        } else { // use bitmap data directly
            // generate texture
            glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE8_ALPHA8,
                          bm.Width(), bm.Height(),
                          0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, map );
        };    
    };    
    
    renderer.Textures->Pop();

    return 0;
};    



__fastcall Aegis::F::Bitmap*
Aegis::G2::Texture::MakeBitmap(const Aegis::G1::Renderer & rn, int mipmap) const {
    int w,h;

    rn.Textures->Push();
    rn.Textures->Enable(0,GL_TEXTURE_2D);
    this->Bind( rn );

    glGetTexLevelParameteriv( GL_TEXTURE_2D, mipmap, GL_TEXTURE_WIDTH, &w );
    glGetTexLevelParameteriv( GL_TEXTURE_2D, mipmap, GL_TEXTURE_HEIGHT, &h );

    Aegis::F::Bitmap *bitmap = new Aegis::F::Bitmap( w, h );
    if (!bitmap->Pointer()) { // not enough memory
        delete bitmap;
        return NULL;
    };    
    
    glGetTexImage( GL_TEXTURE_2D, mipmap, GL_RGBA, GL_UNSIGNED_BYTE,
                   bitmap->Pointer() );
    rn.Textures->Pop();
    
    return bitmap;
};


