/*
  Name: aM1_FastMath.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - M (Math) Subsystem - M1 - Set of fast math functions
*/

#include "aM1_FastMath.hpp"
#include "aM1_CommonMath.hpp"

#include <stdio.h>
#include <math.h>

namespace Aegis {
    namespace M1 {
        void Aegis::M1::FastMathInitialize(); // internally called by FastMathEnable
        
        static bool bFastMathInitialized = false;
        static bool bFastMathEnabled = false;
        
        static unsigned int fast_sqrt_table[0x10000]; // 256k of memory required
        static float fast_acos_table[32768]; // 128k of memory
        static float fast_cos_table[32768]; // 128k of memory
    };
};

void Aegis::M1::FastMathInitialize() {
    // SQRT Table
    unsigned int i;
    union {
        float f;
        unsigned int i;
    } s;
  
    for (i = 0; i <= 0x7FFF; ++i) {
        s.i = (i << 8) | (0x7F << 23);
        s.f = (float) ::sqrtf( s.f ); // precompute using "good" sqrtf
    
        fast_sqrt_table[i + 0x8000] = (s.i & 0x7FFFFF);
    
        s.i = (i << 8) | (0x80 << 23);
        s.f = (float) ::sqrtf( s.f );
    
        fast_sqrt_table[i] = (s.i & 0x7FFFFF);
    }
    
    // COS (and SIN) table
    for (i = 0; i < 32767; ++i) {
        float x = (i * Aegis::M1::PI)/16383.0f;
        fast_cos_table[i] = ::cosf(x);
    };
    fast_cos_table[32766] = 0.0; // little accuracy fix
    
    // ACOS Table
    for (i = 0; i < 32767; ++i) {
        float x = 1.0f - i/16383.0f;
        fast_acos_table[i] = ::acosf(x);
    };    
    
    return;
};

__fastcall bool Aegis::M1::FastMathEnable( bool enable ) {
    bool olden = bFastMathEnabled;
    bFastMathEnabled = enable;
    
    if (enable && !bFastMathInitialized) {
        bFastMathInitialized = true;
        FastMathInitialize();
    };    
    
    return olden;
};

__fastcall __pure float Aegis::M1::sqrt( float f ) {
    if (f == 1.0f) return 1.0f;
    if (f == 0.0f) return 0.0f;
    
    if (bFastMathEnabled) {
        if ( 0 == *(unsigned int *)&f ) return 0.0f;
        
        // Something is bad with this function
        // It is two times faster than sqrtf and gives similar results
        *(unsigned int *)&f = fast_sqrt_table[ (*(unsigned int *)&f >> 8) & 0xFFFF ]
                            | ((((*(unsigned int *)&f - 0x3F800000) >> 1)
                            + 0x3F800000) & 0x7F800000);
        return f;
    } else
        return ::sqrtf(f);
};    

__fastcall float Aegis::M1::acos( float f ) {
    if (bFastMathEnabled) {
        // four times faster table lookup
        return fast_acos_table[ Aegis::M1::ftoi((1.0f - f) * 16383.0f) ];
    } else
        return ::acosf(f);
};    

__fastcall float Aegis::M1::inv ( float f ) {
    if (bFastMathEnabled) {
        int i = 2 * 0x3F800000 - *(int *)&f;
        
        float r = *(float *)&i;
        r = r * (2.0f - f * r);
        
        return r;
    } else
        return 1.0f/f;
};    

__fastcall float Aegis::M1::cos( float f ) {
    if (bFastMathEnabled) {
        register float PI2 = 2.0f * Aegis::M1::PI;
        
        while (f < 0.0) f += PI2;
        while (f > PI2) f -= PI2;

        return fast_cos_table[ Aegis::M1::ftoi(f/(Aegis::M1::PI/16383.0f)) ];
    } else
        return ::cosf(f);
};    

__fastcall int Aegis::M1::ftoi( float f ) {
    if (bFastMathEnabled) {
       int i = 0;
       
//       i = (int)f;
       __asm volatile (
          "fld (%0)   \n\t"
          "frndint    \n\t"
          "fistp (%1) \n\t"
          :
          : "r" (&f), "r" (&i) );
//*/
        return i;
    } else
        return (int)f;
};    


/* // DEPRECATED

__fastcall __pure float Aegis::M1::sin( float f ) {
    if (bFastMathEnabled) {
        f -= 0.5f * Aegis::M1::PI;

        register float PI2 = 2.0f * Aegis::M1::PI;
        
        while (f < 0.0) f += PI2;
        while (f > PI2) f -= PI2;

        return fast_cos_table[(int)(f/(Aegis::M1::PI/16383.0f))];
    } else
        return ::sinf(f);
};    

*/






