/*
  Name: aM2_Color.HPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - M (Math) Subsystem - M2 - Color Class
*/

#ifndef _AM2_COLOR_HPP
#define _AM2_COLOR_HPP

#include <windows.h>

namespace Aegis {
    namespace M2 {
        
        struct Color;
        
    };
};

#include "../../Attribs.hpp"
#include "../M1/aM1_FastMath.hpp"
#include "aM2_RGBA.hpp"

namespace Aegis {
    namespace M2 {
        
        struct __aligned __packed Color {
            union {
                struct {
                    float r;
                    float g;
                    float b;
                    float a;
                };
                float c[4];
            };
            
            Color() : r(0.0f), g(0.0f), b(0.0f), a(1.0f) { };
            inline Color( float ir, float ig, float ib, float ia = 1.0 ) :
                          r(ir), g(ig), b(ib), a(ia) { };
            inline Color( const Color& n ) : r(n.r), g(n.g), b(n.b), a(n.a) { };
            inline Color( COLORREF c ) : r(GetRValue(c)), g(GetGValue(c)),
                                         b(GetBValue(c)), a(1.0) { };
            inline Color( const RGBA& n ) : r(n.red   / 255.0f),
                                            g(n.green / 255.0f),
                                            b(n.blue  / 255.0f),
                                            a(n.alpha / 255.0f) { };

            inline operator COLORREF() { return RGB( (unsigned char)(255*r),
                                                     (unsigned char)(255*g),
                                                     (unsigned char)(255*b) ); };
            inline operator Aegis::M2::RGBA() {
                return RGBA( (unsigned char)(255*r),
                             (unsigned char)(255*g),
                             (unsigned char)(255*b),
                             (unsigned char)(255*a) );
//                return x;
            };
            
            // access operators, non checked boundaries
            inline float& operator[](const int n) { return c[n]; };
            inline const float& operator[](const int n) const { return c[n]; };
            
            // some useful operators
            inline void operator+=(const Color& n){ r += n.r; g += n.g; b += n.b; a += n.a; }
            inline void operator-=(const Color& n){ r -= n.r; g -= n.g; b -= n.b; a -= n.a; }
            inline void operator*=(const Color& n){ r *= n.r; g *= n.g; b *= n.b; a *= n.a; }
            inline void operator/=(const Color& n){ r /= n.r; g /= n.g; b /= n.b; a /= n.a; }
            inline void operator+=(const float f){ r += f; g += f; b += f;/* a += f; */}
            inline void operator-=(const float f){ r -= f; g -= f; b -= f;/* a -= f; */}
            inline void operator*=(const float f){ r *= f; g *= f; b *= f;/* a *= f; */}
            inline void operator/=(const float f){ r /= f; g /= f; b /= f;/* a /= f; */}
            
            // some modification functions
            __fastcall void Correct(); // clamps values to 0.0-1.0 range
            __fastcall void PreMultiply(); // premultiplies color by alpha
        };
        
        // useful operators
        inline Color operator+ ( const Color& a, const Color& b ) { return Color( a.r+b.r, a.g+b.g, a.b+b.b, a.a+b.a ); }
        inline Color operator- ( const Color& a, const Color& b ) { return Color( a.r-b.r, a.g-b.g, a.b-b.b, a.a-b.a ); }
        inline Color operator* ( const Color& a, const Color& b ) { return Color( a.r*b.r, a.g*b.g, a.b*b.b, a.a*b.a ); }
        inline Color operator/ ( const Color& a, const Color& b ) { return Color( a.r/b.r, a.g/b.g, a.b/b.b, a.a/b.a ); }
        
        inline Color operator+ ( const float f, const Color& a ) { return Color( f+a.r, f+a.g, f+a.b, f+a.a ); }
        inline Color operator- ( const float f, const Color& a ) { return Color( f-a.r, f-a.g, f-a.b, f-a.a ); }
        inline Color operator* ( const float f, const Color& a ) { return Color( f*a.r, f*a.g, f*a.b, f*a.a ); }
        inline Color operator+ ( const Color& a, const float f ) { return Color( a.r+f, a.g+f, a.b+f, a.a+f ); }
        inline Color operator- ( const Color& a, const float f ) { return Color( a.r-f, a.g-f, a.b-f, a.a-f ); }
        inline Color operator* ( const Color& a, const float f ) { return Color( a.r*f, a.g*f, a.b*f, a.a*f ); }
        inline Color operator/ ( const Color& a, const float f ) { register float f_ = Aegis::M1::inv(f);
                                                                   return Color( a.r*f_, a.g*f_, a.b*f_, a.a*f_ ); }
        
        inline bool operator< ( const Color& a, const Color& b ) { return ((a.r <  b.r) && (a.g <  b.g) && (a.b <  b.b) && (a.a <  b.a)); }
        inline bool operator<=( const Color& a, const Color& b ) { return ((a.r <= b.r) && (a.g <= b.g) && (a.b <= b.b) && (a.a <= b.a)); }
        inline bool operator> ( const Color& a, const Color& b ) { return ((a.r >  b.r) && (a.g >  b.g) && (a.b >  b.b) && (a.a >  b.a)); }
        inline bool operator>=( const Color& a, const Color& b ) { return ((a.r >= b.r) && (a.g >= b.g) && (a.b >= b.b) && (a.a >= b.a)); }
        inline bool operator==( const Color& a, const Color& b ) { return ((a.r == b.r) && (a.g == b.g) && (a.b == b.b) && (a.a == b.a)); }
        inline bool operator!=( const Color& a, const Color& b ) { return ((a.r != b.r) || (a.g != b.g) || (a.b != b.b) || (a.a != b.a)); }
        
    };
};

#endif

