/*
  Name: aM2_Normal.CPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - M (Math) Subsystem - M2 - Normal Class
*/

#include "aM2_Normal.hpp"
#include "../M1/aM1_CommonMath.hpp"
#include "../M1/aM1_FastMath.hpp"
#include <math.h>

__fastcall Aegis::M2::Normal::Normal( const Vertex& v1, const Vertex& v2 ) {
    
    this->x = v2.x - v1.x;
    this->y = v2.y - v1.y;
    this->z = v2.z - v1.z;
    
    return;
};

__fastcall Aegis::M2::Normal::Normal( const Vertex* face ) {

    // lets hope that no temporary will be created
    register Aegis::M2::Normal tn1( face[2], face[0] );
    register Aegis::M2::Normal tn2( face[1], face[0] );
    
    *this = Cross( tn1, tn2 );
    this->Normalize();
    
    return;
};

__fastcall Aegis::M2::Normal::Normal( const Vertex& v1, const Vertex& v2,
                                      const Vertex& v3 ) {
    register Aegis::M2::Normal tn1( v3, v1 );
    register Aegis::M2::Normal tn2( v2, v1 );
    
    *this = Cross( tn1, tn2 );
    this->Normalize();
    
    return;
};

__fastcall float Aegis::M2::Normal::Length() const {
    return Aegis::M1::sqrt( (this->x * this->x) + 
                            (this->y * this->y) +
                            (this->z * this->z) );
};
    
__fastcall void Aegis::M2::Normal::Normalize() {
    register float mg = this->Length();
    
    if (mg == 1.0f) return;
    if (mg == 0.0f) return;
    
    *this /= mg;
/*    this->x /= mg;
    this->y /= mg;
    this->z /= mg;//*/
	
	return;
};    

__fastcall Aegis::M2::Normal Aegis::M2::Normal::Normalized() const {
    register float mg = this->Length();

    // copy this normal (may be a bit faster) if no division required
    if (mg == 1.0) return Aegis::M2::Normal(*this);
    if (mg == 0.0) return Aegis::M2::Normal();

//    return Aegis::M::Normal( this->x / mg, this->y / mg, this->z / mg );
    return *this / mg;
};    

__fastcall void Aegis::M2::Normal::Reflect( const Normal& normal ) {
    // this normal gets reflected by "normal" of the plane
    
	*this -= 2.0f * normal * Aegis::M2::Dot( *this, normal );
    return;
};

__fastcall void Aegis::M2::Normal::RotateX( float angle ) {
	float s = sinf( Aegis::M1::PI_180 * angle );
	float c = cosf( Aegis::M1::PI_180 * angle );
	
	float Y = this->y;
	this->y = this->y*c - this->z*s;
	this->z = Y*s + this->z*c;

    return;
};

__fastcall void Aegis::M2::Normal::RotateY( float angle ) {
	float s = sinf( Aegis::M1::PI_180 * angle );
	float c = cosf( Aegis::M1::PI_180 * angle );
    
    float X = this->x;
    this->x = this->x*c + this->z*s;
    this->z = -X*s + this->z*c;

    return;
};

__fastcall void Aegis::M2::Normal::RotateZ( float angle ) {
	float s = sinf( Aegis::M1::PI_180 * angle );
	float c = cosf( Aegis::M1::PI_180 * angle );
    
    float X = this->x;
    this->x = this->x*c - this->y*s;
    this->y = X*s + this->y*c;

    return;
};

__fastcall void Aegis::M2::Normal::Rotate( const Normal& axis, float angle ) {
    float costheta, sintheta;
    Normal oldthis = *this, r = axis;

	angle *= Aegis::M1::PI_180;
	
	r.Normalize();
	costheta = (float) Aegis::M1::cos( angle );
	sintheta = (float) Aegis::M1::cos( angle - Aegis::M1::PI_4 );

	this->x  = (costheta + (1.0f - costheta) * r.x * r.x) * oldthis.x;
	this->x += ((1.0f - costheta) * r.x * r.y - r.z * sintheta) * oldthis.y;
	this->x += ((1.0f - costheta) * r.x * r.z + r.y * sintheta) * oldthis.z;

	this->y  = ((1.0f - costheta) * r.x * r.y + r.z * sintheta) * oldthis.x;
	this->y += (costheta + (1.0f - costheta) * r.y * r.y) * oldthis.y;
	this->y += ((1.0f - costheta) * r.y * r.z - r.x * sintheta) * oldthis.z;

	this->z  = ((1.0f - costheta) * r.x * r.z - r.y * sintheta) * oldthis.x;
	this->z += ((1.0f - costheta) * r.y * r.z + r.x * sintheta) * oldthis.y;
	this->z += (costheta + (1.0f - costheta) * r.z * r.z) * oldthis.z;
    
    return;
};

__fastcall void Aegis::M2::Normal::Pack01() {
    this->Normalize();
    *this *= 0.5f;
    *this += 0.5f;
    
    return;
};    

__fastcall Aegis::M2::Normal Aegis::M2::Normal::Packed01() const {
    Aegis::M2::Normal nn = this->Normalized();
    
    nn *= 0.5f;
    nn += 0.5f;
    
    return nn;
};    

    
// another functions
__fastcall Aegis::M2::Normal Aegis::M2::Cross( const Normal& n1,
                                               const Normal& n2 ) {
    float x = (n1.y * n2.z) - (n1.z * n2.y);
    float y = (n1.z * n2.x) - (n1.x * n2.z);
    float z = (n1.x * n2.y) - (n1.y * n2.x);

    return Aegis::M2::Normal(x,y,z);
};

__fastcall float Aegis::M2::Dot( const Normal& n1, const Normal& n2 ) {
    return (n1.x * n2.x) + (n1.y * n2.y) + (n1.z * n2.z);
};

__fastcall float Aegis::M2::Angle( const Normal& n1, const Normal& n2 ) {
    return( Aegis::M1::acos( Dot( n1, n2 ) / (n1.Length() * n2.Length()) ) );
};

/*
__fastcall Normal InterpolateLinear( Normal& n1, Normal& n2, float x ) {
	Aegis::M::Normal m, n(n1);
	
	float angle = Aegis::M::Angle( a, b );
	m.Cross( a, b );
	
	n.ArbitraryRotate( angle * t / PI180, m );

	return n;
};    
*/





