/*
  Name: aM2_Normal.HPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - M (Math) Subsystem - M2 - Normal Class
*/

#ifndef _AM2_NORMAL_HPP
#define _AM2_NORMAL_HPP

#include <stdio.h>

namespace Aegis {
    namespace M2 {
        class Normal;
    };
};

#include "../../Attribs.hpp"
#include "aM2_Vertex.hpp"
#include "../M1/aM1_FastMath.hpp"
#include "../M1/aM1_CommonMath.hpp"

namespace Aegis {
    namespace M2 {
        
        class __aligned __packed Normal {
            public:
                
            union {
                struct {
                    float x;
                    float y;
                    float z;
                };
                float v[4]; // 4 - to be more like Vertex, Color and TexCoord
            };
            
            // normal constructors
            inline Normal() : x(0.0f), y(0.0f), z(0.0f) { };
            inline Normal( float ix, float iy, float iz ) : x(ix), y(iy), z(iz) { };
            inline Normal( const Normal& n ) : x(n.x), y(n.y), z(n.z) { };
            inline Normal( const Vertex& v ) : x(v.x), y(v.y), z(v.z) { };
            inline Normal( float angle ) : x( Aegis::M1::cos( (2.0f * Aegis::M1::PI * (angle - 90.0f)) / 360.0f ) ),
                                           y( Aegis::M1::cos( (2.0f * Aegis::M1::PI * (angle +  0.0f)) / 360.0f ) ),
                                           z( 0.0f ) { };

            // creates normal from v1 facing to v2
            __fastcall Normal( const Vertex& v1, const Vertex& v2 );
            
            // creates normal for given face
            __fastcall Normal( const Vertex* face );
            __fastcall Normal( const Vertex& v1, const Vertex& v2,
                               const Vertex& v3 );
                               
            // builders
            inline void Construct( float ix, float iy, float iz ) {
                this->x = ix;
                this->y = iy;
                this->z = iz;
                
                return;
            };
            inline void Zero() { x = 0.0f; y = 0.0f; z = 0.0f; };
            inline void One() { x = 1.0f; y = 1.0f; z = 1.0f; };
            
            // access operators, non checked boundaries
            inline float& operator[](const int n) { return v[n]; };
            inline const float& operator[](const int n) const { return v[n]; };
            
            // to vertex
            inline operator Vertex() const { return Vertex(x, y, z); };
            
            // useful operators
            inline Normal operator+() const { return *this; };
            inline Normal operator-() const { return Normal( -x, -y, -z); };

            // normal validity operators
            inline operator bool() const { return ((x != 0.0f) || (y != 0.0f) || (z != 0.0f)); }
//            inline bool operator!() const { return ((x == 0.0f) && (y == 0.0f) && (z == 0.0f)); }
            
            inline void operator+=(const Normal& a){ x += a.x; y += a.y; z += a.z; }
            inline void operator-=(const Normal& a){ x -= a.x; y -= a.y; z -= a.z; }
            inline void operator*=(const Normal& a){ x *= a.x; y *= a.y; z *= a.z; }
            inline void operator/=(const Normal& a){ x /= a.x; y /= a.y; z /= a.z; }
            inline void operator+=(const float f){ x += f; y += f; z += f; }
            inline void operator-=(const float f){ x -= f; y -= f; z -= f; }
            inline void operator*=(const float f){ x *= f; y *= f; z *= f; }
            inline void operator/=(const float f){ x /= f; y /= f; z /= f; }
            
            // some functions
            __fastcall float Length() const;
            __fastcall void Normalize();
            __fastcall Aegis::M2::Normal Normalized() const;
            
            __fastcall void Reflect( const Normal& normal );

        	__fastcall void RotateX( float angle );
        	__fastcall void RotateY( float angle );
        	__fastcall void RotateZ( float angle );
        	__fastcall void Rotate( const Normal& axis, float angle );
    
            __fastcall void Pack01();
            __fastcall Aegis::M2::Normal Packed01() const;
        };
        
        // useful operators
        inline Normal operator+ ( const Normal& a, const Normal& b ) { return Normal( a.x+b.x, a.y+b.y, a.z+b.z ); }
        inline Normal operator- ( const Normal& a, const Normal& b ) { return Normal( a.x-b.x, a.y-b.y, a.z-b.z ); }
        inline Normal operator* ( const Normal& a, const Normal& b ) { return Normal( a.x*b.x, a.y*b.y, a.z*b.z ); }
        inline Normal operator/ ( const Normal& a, const Normal& b ) { return Normal( a.x/b.x, a.y/b.y, a.z/b.z ); }
        
        inline Normal operator+ ( const float f, const Normal& a ) { return Normal( f+a.x, f+a.y, f+a.z ); }
        inline Normal operator- ( const float f, const Normal& a ) { return Normal( f-a.x, f-a.y, f-a.z ); }
        inline Normal operator* ( const float f, const Normal& a ) { return Normal( f*a.x, f*a.y, f*a.z ); }
        inline Normal operator+ ( const Normal& a, const float f ) { return Normal( a.x+f, a.y+f, a.z+f ); }
        inline Normal operator- ( const Normal& a, const float f ) { return Normal( a.x-f, a.y-f, a.z-f ); }
        inline Normal operator* ( const Normal& a, const float f ) { return Normal( a.x*f, a.y*f, a.z*f ); }
        inline Normal operator/ ( const Normal& a, const float f ) { register float f_ = Aegis::M1::inv(f);
                                                                     return Normal( a.x*f_, a.y*f_, a.z*f_ ); }
        
        inline bool operator< ( const Normal& a, const Normal& b ) { return ((a.x <  b.x) && (a.y <  b.y) && (a.z <  b.z)); }
        inline bool operator<=( const Normal& a, const Normal& b ) { return ((a.x <= b.x) && (a.y <= b.y) && (a.z <= b.z)); }
        inline bool operator> ( const Normal& a, const Normal& b ) { return ((a.x >  b.x) && (a.y >  b.y) && (a.z >  b.z)); }
        inline bool operator>=( const Normal& a, const Normal& b ) { return ((a.x >= b.x) && (a.y >= b.y) && (a.z >= b.z)); }
        inline bool operator==( const Normal& a, const Normal& b ) { return ((a.x == b.x) && (a.y == b.y) && (a.z == b.z)); }
        inline bool operator!=( const Normal& a, const Normal& b ) { return ((a.x != b.x) || (a.y != b.y) || (a.z != b.z)); }
        
        
        // good functions
        __fastcall Normal Cross( const Normal& n1, const Normal& n2 );
        __fastcall float Dot( const Normal& n1, const Normal& n2 );
        __fastcall float Angle( const Normal& n1, const Normal& n2 );

//        __fastcall Normal InterpolateLinear( Normal& n1, Normal& n2, float x );
       
       
        // type alias
        typedef Normal Vector;
    };
};

#endif

