/*
  Name: aM2_RGBA.HPP
  Author: Jan Ringo, Tringi@Mx-3.cz, http://Tringi.Mx-3.cz
  Description: Aegis Engine - M (Math) Subsystem - M2 - RGBA Class
*/

#ifndef _AM2_RGBA_HPP
#define _AM2_RGBA_HPP

#include <windows.h>
#include "../../Attribs.hpp"
#include "../M1/aM1_FastMath.hpp"

namespace Aegis {
    namespace M2 {
        
        struct __packed RGBA {
            union {
                struct __packed {
                    union {
                        unsigned char red;
                        unsigned char r;
                        signed char red_offset;
                    };
                    union {
                        unsigned char green;
                        unsigned char g;
                        signed char green_offset;
                    };
                    union {
                        unsigned char blue;
                        unsigned char b;
                        signed char blue_offset;
                    };
                    union {
                        unsigned char alpha;
                        unsigned char a;
                        signed char alpha_offset;
                    };
                };
                
                unsigned char color[4];
                signed char   offset[4];

                DWORD dwColor;
                COLORREF crColor;
            };
            
            RGBA() : red(0), green(0), blue(0), alpha(255) { };
            inline RGBA( unsigned char nr, unsigned char ng,
                         unsigned char nb, unsigned char na = 255) :
                         red(nr), green(ng), blue(nb), alpha(na) { };

            inline RGBA( const RGBA& n ) : red(n.red), green(n.green),
                                           blue(n.blue), alpha(n.alpha) { };
            inline RGBA( COLORREF c ) : red(GetRValue(c)), green(GetGValue(c)),
                                        blue(GetBValue(c)), alpha(255) { };

            inline operator COLORREF() { return RGB( (unsigned char) red,
                                                     (unsigned char) green, 
                                                     (unsigned char) blue ); };
            
            // access operators, non checked boundaries
            inline unsigned char& operator[](const int n) { return color[n]; };
            inline const unsigned char& operator[](const int n) const
                                                          { return color[n]; };
            
            // some useful operators
            inline void operator+=(const RGBA& n){ red += n.red; green += n.green; blue += n.blue; alpha += n.alpha; }
            inline void operator-=(const RGBA& n){ red -= n.red; green -= n.green; blue -= n.blue; alpha -= n.alpha; }
            inline void operator*=(const RGBA& n){ red *= n.red; green *= n.green; blue *= n.blue; alpha *= n.alpha; }
            inline void operator/=(const RGBA& n){ red /= n.red; green /= n.green; blue /= n.blue; alpha /= n.alpha; }
            inline void operator+=(const unsigned char f){ red += f; green += f; blue += f;/* alpha += f; */}
            inline void operator-=(const unsigned char f){ red -= f; green -= f; blue -= f;/* alpha -= f; */}
            inline void operator*=(const unsigned char f){ red *= f; green *= f; blue *= f;/* alpha *= f; */}
            inline void operator/=(const unsigned char f){ red /= f; green /= f; blue /= f;/* alpha /= f; */}
            
            // some modification functions
            __fastcall void PreMultiply(); // premultiplies color by alpha

        };
        
        // useful operators
        inline RGBA operator+ ( const RGBA& a, const RGBA& b ) { return RGBA( a.r+b.r, a.g+b.g, a.b+b.b, a.a+b.a ); }
        inline RGBA operator- ( const RGBA& a, const RGBA& b ) { return RGBA( a.r-b.r, a.g-b.g, a.b-b.b, a.a-b.a ); }
        inline RGBA operator* ( const RGBA& a, const RGBA& b ) { return RGBA( a.r*b.r, a.g*b.g, a.b*b.b, a.a*b.a ); }
        inline RGBA operator/ ( const RGBA& a, const RGBA& b ) { return RGBA( a.r/b.r, a.g/b.g, a.b/b.b, a.a/b.a ); }
        
        inline RGBA operator+ ( const unsigned char f, const RGBA& a ) { return RGBA( f+a.r, f+a.g, f+a.b, f+a.a ); }
        inline RGBA operator- ( const unsigned char f, const RGBA& a ) { return RGBA( f-a.r, f-a.g, f-a.b, f-a.a ); }
        inline RGBA operator* ( const unsigned char f, const RGBA& a ) { return RGBA( f*a.r, f*a.g, f*a.b, f*a.a ); }
        inline RGBA operator+ ( const RGBA& a, const unsigned char f ) { return RGBA( a.r+f, a.g+f, a.b+f, a.a+f ); }
        inline RGBA operator- ( const RGBA& a, const unsigned char f ) { return RGBA( a.r-f, a.g-f, a.b-f, a.a-f ); }
        inline RGBA operator* ( const RGBA& a, const unsigned char f ) { return RGBA( a.r*f, a.g*f, a.b*f, a.a*f ); }
        inline RGBA operator/ ( const RGBA& a, const unsigned char f ) { return RGBA( a.r/f, a.g/f, a.b/f, a.a/f ); }
        
        inline bool operator< ( const RGBA& a, const RGBA& b ) { return ((a.r <  b.r) && (a.g <  b.g) && (a.b <  b.b) && (a.a <  b.a)); }
        inline bool operator<=( const RGBA& a, const RGBA& b ) { return ((a.r <= b.r) && (a.g <= b.g) && (a.b <= b.b) && (a.a <= b.a)); }
        inline bool operator> ( const RGBA& a, const RGBA& b ) { return ((a.r >  b.r) && (a.g >  b.g) && (a.b >  b.b) && (a.a >  b.a)); }
        inline bool operator>=( const RGBA& a, const RGBA& b ) { return ((a.r >= b.r) && (a.g >= b.g) && (a.b >= b.b) && (a.a >= b.a)); }
        inline bool operator==( const RGBA& a, const RGBA& b ) { return ((a.r == b.r) && (a.g == b.g) && (a.b == b.b) && (a.a == b.a)); }
        inline bool operator!=( const RGBA& a, const RGBA& b ) { return ((a.r != b.r) || (a.g != b.g) || (a.b != b.b) || (a.a != b.a)); }
        
    };
};

#endif

